import copy
import math
import os
import shutil
import time
import unittest
from multiprocessing.sharedctypes import Value

import cv2
import geomapi.utils.geometryutils as gt
import ifcopenshell
import numpy as np
import open3d as o3d
import pye57
import pytest
from ifcopenshell.util.selector import Selector
from rdflib import Graph

 ################################## SETUP/TEARDOWN MODULE ######################

# def setUpModule():
#     #execute once before the module 
#     print('-----------------Setup Module----------------------')

# def tearDownModule():
#     #execute once after the module 
#     print('-----------------TearDown Module----------------------')

class TestGeometryutils(unittest.TestCase):

 ################################## SETUP/TEARDOWN CLASS ######################
  
    @classmethod
    def setUpClass(cls):
        #execute once before all tests
        print('-----------------Setup Class----------------------')
        st = time.time()
        cls.path=os.path.join(os.getcwd(), "test","testfiles") #os.pardir, 
        
        #POINTCLOUD1
        cls.e57Path1=os.path.join(cls.path,'PCD',"week22 photogrammetry - Cloud.e57")     
        #POINTCLOUD2
        cls.e57Path2=os.path.join(cls.path,'PCD',"week 22 - Lidar.e57")     
       
        #POINTCLOUD3
        cls.pcdPath1=os.path.join(cls.path,'PCD',"academiestraat week 22 a 20.pcd")     
        cls.pcd= o3d.io.read_point_cloud(cls.pcdPath1) 

        #MESH
        cls.meshPath=os.path.join(cls.path,'MESH',"week22.obj")  
        cls.mesh= o3d.io.read_triangle_mesh(cls.meshPath)
    
        #IMG
        cls.image1Path=os.path.join(cls.path,'IMG',"IMG_2173.JPG")  
        cls.image1=cv2.imread(cls.image1Path)
        cls.image1CartesianTransform= np.array([[-0.0544245051021791, 0.089782180920334, 0.994473294379276, -8.94782374621677],
                                                [-0.783686718502243, -0.621016494566922, 0.0131772804097903 ,11.2531401937057],
                                                [0.618767404189627, -0.778638345745315, 0.104159618122526, 6.5428452363933],
                                                [0,0,0,1]])
        cls.image2Path=os.path.join(cls.path,'IMG',"IMG_2174.JPG")  
        cls.image2=cv2.imread(cls.image1Path)
        cls.image2CartesianTransform= np.array([[-0.046509031201878, 0.0485391010476459, 0.99773787423659, -8.63657982153356],
                                                [-0.714937318920729, -0.699188212147553, 0.00068848264200394 ,9.21354145067747],
                                                [0.697639978807912, -0.713288020131696, 0.0672209811405822,6.57082854991429],
                                                [0,0,0,1]])
        #IFC1
        cls.ifcPath1=os.path.join(cls.path,'IFC',"Academiestraat_building_1.ifc") 
        cls.classes= '.IfcBeam | .IfcColumn | .IfcWall | .IfcSlab'
        ifc1 = ifcopenshell.open(cls.ifcPath1)   
        selector = Selector() 
        cls.bimMeshes=[gt.ifc_to_mesh(ifcElement) for ifcElement in selector.parse(ifc1, cls.classes)]     
        cls.bimBoxes=[mesh.get_oriented_bounding_box() for mesh in cls.bimMeshes if mesh]

        #IFC2
        cls.ifcPath2=os.path.join(cls.path,'IFC',"Academiestraat_parking.ifc") 
        ifc2 = ifcopenshell.open(cls.ifcPath2)   
        ifcSlab=ifc2.by_guid('2qZtnImXH6Tgdb58DjNlmF')
        ifcWall=ifc2.by_guid('06v1k9ENv8DhGMCvKUuLQV')
        ifcBeam=ifc2.by_guid('05Is7PfoXBjhBcbRTnzewz' )
        ifcColumn=ifc2.by_guid('23JN72MijBOfF91SkLzf3a')
        # ifcWindow=ifc.by_guid(cls.slabGlobalid) 
        # ifcDoor=ifc.by_guid(cls.slabGlobalid)

        cls.slabMesh=gt.ifc_to_mesh(ifcSlab)
        cls.wallMesh=gt.ifc_to_mesh(ifcWall)
        cls.beamMesh=gt.ifc_to_mesh(ifcBeam)
        cls.columnMesh=gt.ifc_to_mesh(ifcColumn)
        # cls.windowMesh=gt.ifc_to_mesh(ifcWindow)
        # cls.doorMesh=gt.ifc_to_mesh(ifcDoor)

        #RESOURCES
        cls.resourcePath=os.path.join(cls.path,"resources")
        if not os.path.exists(cls.resourcePath):
            os.mkdir(cls.resourcePath)
   
        et = time.time()
        print("startup time: "+str(et - st))
        print('{:50s} {:5s} '.format('tests','time'))
        print('------------------------------------------------------')


    @classmethod
    def tearDownClass(cls):
        #execute once after all tests
        print('-----------------TearDown Class----------------------')
        shutil.rmtree(cls.resourcePath)      

################################## SETUP/TEARDOWN ######################

    def setUp(self):
        #execute before every test
        self.startTime = time.time()   

    def tearDown(self):
        #execute after every test
        t = time.time() - self.startTime
        print('{:50s} {:5s} '.format(self._testMethodName,str(t)))

################################## FIXTURES ######################
    # # @pytest.fixture(scope='module')
    # # @pytest.fixture
    # def test_data(*args):
    #     here = os.path.split(__file__)[0]
    #     return os.path.join(here, "testfiles", *args)

    # @pytest.fixture
    # def e57Path1():
    #     return test_data("pointcloud.e57")

    # @pytest.fixture
    # def ifcData():
    #     ifcPath=os.path.join(os.getcwd(),"testfiles", "ifcfile.ifc")
    #     classes= '.IfcBeam | .IfcColumn | .IfcWall | .IfcSlab'
    #     ifc = ifcopenshell.open(ifcPath)   
    #     selector = Selector()
    #     dataList=[]
    #     for ifcElement in selector.parse(ifc, classes): 
    #         dataList.append(ifcElement)
    #     return dataList

################################## TEST FUNCTIONS ######################

    def test_arrays_to_mesh_and_mesh_to_arrays(self):
        #mesh_to_arrays
        tuple=gt.mesh_to_arrays(self.meshPath)
        self.assertEqual(len(tuple[0]),len(self.mesh.vertices))
        self.assertEqual(len(tuple[1]),len(self.mesh.triangles))
        self.assertEqual(len(tuple[2]),len(self.mesh.vertex_colors))
        self.assertEqual(tuple[3],None)
        self.assertEqual(tuple[4],0)

        #arrays_to_mesh
        mesh=gt.arrays_to_mesh(tuple)
        self.assertEqual(len(mesh.vertices),len(self.mesh.vertices))
        self.assertEqual(len(mesh.triangles),len(self.mesh.triangles))
        self.assertEqual(len(mesh.vertex_colors),len(self.mesh.vertex_colors))

    def test_arrays_to_pcd(self):
        #pcd_to_arrays
        tuple=gt.pcd_to_arrays(self.pcdPath1)
        self.assertEqual(len(tuple[0]),len(self.pcd.points))
        self.assertEqual(len(tuple[1]),len(self.pcd.colors))
        self.assertEqual(tuple[2],None)
        self.assertEqual(tuple[3],0)
     
        #arrays_to_mesh
        pcd=gt.arrays_to_pcd(tuple)
        self.assertEqual(len(pcd.points),len(self.pcd.points))
        self.assertEqual(len(pcd.colors),len(self.pcd.colors))
        self.assertEqual(len(pcd.normals),len(self.pcd.normals))

    def test_create_identity_point_cloud(self):
        # 1 geometry
        identityPointCloud, indentityArray=gt.create_identity_point_cloud(self.slabMesh)
        self.assertEqual(len(identityPointCloud.points),len(indentityArray))

        #multiple geometries
        list=[self.slabMesh, self.wallMesh]
        identityPointCloud2, indentityArray2=gt.create_identity_point_cloud(list)
        self.assertEqual(len(identityPointCloud2.points),len(indentityArray2))

    def test_create_visible_point_cloud_from_meshes(self):
        referenceMesh1= copy.deepcopy(self.slabMesh)
        referenceMesh1.translate([1,0,0])
        referenceMesh2= copy.deepcopy(self.slabMesh)
        referenceMesh2.translate([0,1,0])
        
        # 1 geometry
        identityPointClouds1, percentages1=gt.create_visible_point_cloud_from_meshes(geometries=self.slabMesh,
                                                references=referenceMesh1)
        self.assertEqual(len(identityPointClouds1),1)
        self.assertGreater(len(identityPointClouds1[0].points),10)
        self.assertEqual(len(percentages1),1)        
        self.assertLess(percentages1[0],0.2)

        # multiple geometries 
        list=[self.slabMesh,self.wallMesh]
        references=[referenceMesh1,referenceMesh2]
        identityPointClouds2, percentages2=gt.create_visible_point_cloud_from_meshes(geometries=list,
                                                references=references)
        self.assertEqual(len(identityPointClouds2),2)
        self.assertLess(len(identityPointClouds2[0].points),len(identityPointClouds1[0].points))
        self.assertEqual(len(percentages2),2)        
        self.assertLess(percentages2[0],percentages1[0])
    
        

    def test_cap_mesh(self):
        print('test_cap_mesh NOT IMPLEMENTED')
        self.assertEqual(0,0)
        
    def test_box_to_mesh(self):
        box=o3d.geometry.TriangleMesh.create_box(width=1.0, height=1.0, depth=1.0)
        boundingBox=box.get_oriented_bounding_box()
        mesh =gt.box_to_mesh(boundingBox) 
        self.assertIsInstance(mesh,o3d.geometry.TriangleMesh)
       
    def test_ifc_to_mesh(self):
        ifc = ifcopenshell.open(self.ifcPath1)   
        selector = Selector()
        ifcCounter=0
        meshCounter =0
        for ifcElement in selector.parse(ifc, self.classes): 
            ifcCounter+=1
            mesh=gt.ifc_to_mesh(ifcElement)
            self.assertIsInstance(mesh,o3d.geometry.TriangleMesh)
            if len(mesh.vertices) !=0:
                meshCounter +=1
            if ifcCounter==20:
                break
        self.assertEqual(meshCounter,ifcCounter)
      
    def test_get_oriented_bounding_box(self):     
        #cartesianBounds
        boxGt=self.mesh.get_axis_aligned_bounding_box()
        boxPointsGt=np.asarray(boxGt.get_box_points())
        cartesianBounds=gt.get_cartesian_bounds(self.mesh)
        box=gt.get_oriented_bounding_box(cartesianBounds)
        boxPoints=np.asarray(box.get_box_points())

        for i in range(0,7):
            for j in range(0,2):
                self.assertAlmostEqual(boxPointsGt[i][j],boxPoints[i][j],delta=0.01)

        #orientedBounds
        boxGt=self.mesh.get_oriented_bounding_box()
        boxPointsGt=np.asarray(boxGt.get_box_points())
        box=gt.get_oriented_bounding_box(boxPointsGt)
        boxPoints=np.asarray(box.get_box_points())
        
        for i in range(0,7):
            for j in range(0,2):
                self.assertAlmostEqual(boxPointsGt[i][j],boxPoints[i][j],delta=0.01)

    def test_generate_visual_cone_from_image(self):
        cartesianTransform=np.array([[-4.65090312e-02,  4.85391010e-02,  9.97737874e-01, -8.63657982e+00],
                                    [-7.14937319e-01, -6.99188212e-01,  6.88482642e-04,  9.21354145e+00],
                                    [ 6.97639979e-01, -7.13288020e-01 , 6.72209811e-02,  6.57082855e+00],
                                    [ 0.00000000e+00 , 0.00000000e+00,  0.00000000e+00 , 1.00000000e+00]])
        mesh=gt.generate_visual_cone_from_image(cartesianTransform)
        self.assertIsInstance(mesh,o3d.geometry.TriangleMesh)
      
    def test_get_cartesian_transform(self):
        cartesianBounds=np.array([-1.0,1,-0.5,0.5,-5,-4])       
        translation=np.array([1, 2, 3])
        rotation=np.array([1,0,0,5,2,6,4,7,8])

        #no data
        cartesianTransform=gt.get_cartesian_transform()
        self.assertEqual(cartesianTransform.shape[0],4)
        self.assertEqual(cartesianTransform.shape[1],4)
        self.assertEqual(cartesianTransform[1,1],1)
        self.assertEqual(cartesianTransform[2,3],0)

        #rotation + translation
        cartesianTransform=gt.get_cartesian_transform(rotation=rotation,translation=translation)
        self.assertEqual(cartesianTransform[1,1],2)
        self.assertEqual(cartesianTransform[0,3],1)

        #cartesianBounds
        cartesianTransform=gt.get_cartesian_transform(cartesianBounds=cartesianBounds)
        self.assertEqual(cartesianTransform[1,1],1)
        self.assertEqual(cartesianTransform[2,3],-4.5)
        
    def test_get_oriented_bounds(self):        
        box=self.mesh.get_axis_aligned_bounding_box()
        boxPoints=np.asarray(box.get_box_points())
        cartesianBounds=gt.get_cartesian_bounds(self.mesh)
        boundingPoints=np.asarray(gt.get_oriented_bounds(cartesianBounds)) 

        for i in range(0,7):
            for j in range(0,2):
                self.assertAlmostEqual(boundingPoints[i][j],boxPoints[i][j],delta=0.01)
        
    def test_get_box_inliers(self):
        mesh=self.wallMesh.translate([0,0,3])
        wallBox=mesh.get_oriented_bounding_box()

        wallInliers= gt.get_box_inliers(sourceBox=wallBox, testBoxes=self.bimBoxes) 
        self.assertEqual(len(wallInliers),8)
        
    def test_get_box_intersections(self):
        mesh=self.wallMesh.translate([0,0,3])
        wallBox=mesh.get_oriented_bounding_box()
        wallInliers= gt.get_box_intersections(sourceBox=wallBox, testBoxes=self.bimBoxes)
        self.assertEqual(len(wallInliers),41)
        
    def test_get_cartesian_bounds(self):
        #box
        box=self.mesh.get_oriented_bounding_box()
        minBounds=box.get_min_bound()
        maxBounds=box.get_max_bound()
        cartesianBounds=gt.get_cartesian_bounds(box)
        self.assertEqual(minBounds[0],cartesianBounds[0])
        self.assertEqual(maxBounds[2],cartesianBounds[5])

        #mesh
        cartesianBounds=gt.get_cartesian_bounds(self.mesh)
        minBounds=self.mesh.get_min_bound()
        maxBounds=self.mesh.get_max_bound()
        self.assertEqual(minBounds[0],cartesianBounds[0])
        self.assertEqual(maxBounds[2],cartesianBounds[5])

        #pointcloud
        cartesianBounds=gt.get_cartesian_bounds(self.pcd)
        minBounds=self.pcd.get_min_bound()
        maxBounds=self.pcd.get_max_bound()
        self.assertEqual(minBounds[0],cartesianBounds[0])
        self.assertEqual(maxBounds[2],cartesianBounds[5])
        
    def test_get_triangles_center(self):
        mesh=self.mesh      
        triangleIndices=[0,1,2]    
        centers=gt.get_triangles_center(mesh,triangleIndices)
        self.assertEqual(centers.size,9)
        self.assertAlmostEqual(centers[0][0],4.48127794,delta=0.01)
        self.assertAlmostEqual(centers[2][2],5.12661028,delta=0.01)
        
    def test_mesh_to_pcd(self):
        pcd=gt.mesh_to_pcd(self.mesh)
        self.assertIsInstance(pcd,o3d.geometry.PointCloud)
        self.assertGreater(len(pcd.points),3)
        
    def test_generate_virtual_images(self):
        cartesianTransforms= [self.image1CartesianTransform,self.image2CartesianTransform]
        image=gt.generate_virtual_images(self.mesh, cartesianTransforms)
        self.assertEqual(len(image),2)
        self.assertIsInstance(image[0],o3d.cpu.pybind.geometry.Image)
        
    def test_generate_virtual_image(self):
        width=640
        height=480
        fx=400
        fy=400
        cx=width/2-0.5
        cy=height/2-0.5
        pinholeCamera=o3d.camera.PinholeCameraParameters
        pinholeCamera.extrinsic=np.linalg.inv(self.image1CartesianTransform)
        pinholeCamera.intrinsic=o3d.camera.PinholeCameraIntrinsic(width,height,fx,fy,cx,cy)
        image=gt.generate_virtual_image(self.mesh, pinholeCamera)
        self.assertIsInstance(image,o3d.cpu.pybind.geometry.Image)
        
    def test_e57path_to_pcd(self):
        e57=pye57.E57(self.e57Path1) 
        header = e57.get_header(0)
        pcd=gt.e57path_to_pcd(e57Path=self.e57Path1, e57Index=0) 
        self.assertEqual(len(pcd.points) , header.point_count)

    def test_e57_to_arrays(self):
        e57=pye57.E57(self.e57Path1) 
        header = e57.get_header(0)
        tuple=gt.e57_to_arrays(self.e57Path1)
        self.assertEqual(len(tuple),5)
        self.assertEqual(len(tuple[0]),header.point_count)
        self.assertEqual(len(tuple[1]),header.point_count)
        self.assertEqual(len(tuple[2]),header.point_count)

    def test_e57_to_pcd(self):
        e57=pye57.E57(self.e57Path1) 
        header = e57.get_header(0)
        pcd=gt.e57_to_pcd(e57, percentage=0.5)
        self.assertEqual(len(pcd.points),int(header.point_count*0.5))

        e57=pye57.E57(self.e57Path2) 
        header = e57.get_header(1)
        pcd=gt.e57_to_pcd(e57, percentage=0.5)
        self.assertLess(len(pcd.points),6000000)

    def test_e57path_to_pcds_multiprocessing(self):
        e57=pye57.E57(self.e57Path1) 
        header1 = e57.get_header(0)
        pcds=gt.e57path_to_pcds_multiprocessing(self.e57Path1, percentage=0.5)
        self.assertEqual(len(pcds),1)        
        self.assertEqual(len(pcds[0].points),int(header1.point_count*0.5))

        e57=pye57.E57(self.e57Path2) 
        header1 = e57.get_header(1)
        header2 = e57.get_header(2)
        pcds=gt.e57path_to_pcds_multiprocessing(self.e57Path2, percentage=0.5)
        self.assertEqual(len(pcds[0].points),int(header1.point_count*0.5))
        self.assertEqual(len(pcds[1].points),int(header2.point_count*0.5))

    def test_pcd_to_arrays(self):
        tuple=gt.pcd_to_arrays(self.pcdPath1, percentage=0.5)
        self.assertEqual(len(tuple),4)
        self.assertEqual(len(tuple[0]),int(len(self.pcd.points)*0.5))
        self.assertEqual(len(tuple[1]),int(len(self.pcd.points)*0.5))
        self.assertEqual(tuple[2],None)

    def test_mesh_to_arrays(self):
        tuple=gt.mesh_to_arrays(self.meshPath)
        self.assertEqual(len(tuple),5)
        self.assertEqual(len(tuple[0]),len(self.mesh.vertices))
        self.assertEqual(len(tuple[1]),len(self.mesh.triangles))
        self.assertEqual(len(tuple[2]),len(self.mesh.vertex_colors))
        self.assertEqual(tuple[3],None)

    def test_e57_get_colors(self):
        e57=pye57.E57(self.e57Path1) 
        raw_data = e57.read_scan_raw(0)  
        header = e57.get_header(0)
        self.assertEqual(len(raw_data["cartesianX"]) , header.point_count)

        colors=gt.e57_get_colors(raw_data)
        self.assertEqual(len(colors),len(raw_data["cartesianX"]))
        self.assertEqual(len(colors),header.point_count)
        
    def test_crop_geometry_by_box(self):
        #test point cloud
        box=self.mesh.get_oriented_bounding_box()
        pcd=gt.crop_geometry_by_box(self.pcd, box) 
        self.assertIsInstance(pcd,o3d.geometry.PointCloud)
        self.assertGreater(len(pcd.points),10000000)

        #test mesh
        box=self.pcd.get_oriented_bounding_box()
        mesh=gt.crop_geometry_by_box(self.mesh, box, subdivide = 0)
        self.assertIsInstance(mesh,o3d.geometry.TriangleMesh)
        self.assertGreater(len(mesh.vertices),100000)

    def test_get_mesh_representation(self):
        #mesh
        mesh=gt.get_mesh_representation(self.mesh)
        self.assertEqual(len(mesh.triangles),len(self.mesh.triangles))

        #point cloud
        mesh=gt.get_mesh_representation(self.pcd)
        self.assertGreater(len(mesh.triangles),3)

        #invalid input
        self.assertRaises(ValueError,gt.get_mesh_representation,self.mesh.get_oriented_bounding_box())

    def test_get_mesh_inliers(self):
        #mesh
        sources=[self.slabMesh,self.wallMesh]
        indices=gt.get_mesh_inliers(sources=sources,reference=self.mesh)
        self.assertEqual(len(indices),len(sources))
        
        #pcd
        sources=[self.slabMesh,self.wallMesh]
        indices=gt.get_mesh_inliers(sources=sources,reference=self.pcd)
        self.assertEqual(len(indices),len(sources))

        #invalid sources
        self.assertRaises(ValueError,gt.get_mesh_inliers,self.mesh.get_oriented_bounding_box(),self.mesh.get_oriented_bounding_box())

    def test_expand_box(self):
        cartesianBounds=np.array([0,10,7,8.3,-2,2])       
        box=gt.get_oriented_bounding_box(cartesianBounds)
        #positive expansion
        expandedbox1=gt.expand_box(box,u=5,v=3,w=1)
        self.assertEqual(expandedbox1.extent[0],box.extent[0]+5)
        self.assertEqual(expandedbox1.extent[1],box.extent[1]+3)
        self.assertEqual(expandedbox1.extent[2],box.extent[2]+1)
        
        #negate expansion
        expandedbox2=gt.expand_box(box,u=-1,v=-1,w=-1)
        self.assertEqual(expandedbox2.extent[0],box.extent[0]-1)
        self.assertEqual(expandedbox2.extent[1],box.extent[1]-1)
        self.assertEqual(expandedbox2.extent[2],box.extent[2]-1)

        #impossible expansion
        self.assertRaises(ValueError,gt.expand_box,box=box,u=-55 )
            
    def test_join_geometries(self):
        #TriangleMesh
        mesh1=self.mesh
        mesh2=copy.deepcopy(mesh1)
        mesh2.translate([5,0,0])
        joinedMeshes= gt.join_geometries([mesh1,mesh2]) 
        self.assertEqual(len(joinedMeshes.triangles), (len(mesh1.triangles)+len(mesh2.triangles)))
        self.assertTrue(joinedMeshes.has_vertex_colors())

        #PointCloud
        pcd1=self.pcd
        pcd2=copy.deepcopy(pcd1)
        pcd2.translate([5,0,0])
        joinedpcd= gt.join_geometries([pcd1,pcd2]) 
        self.assertEqual(len(joinedpcd.points), (len(pcd1.points)+len(pcd2.points)))
        self.assertTrue(joinedpcd.has_colors())
        
    def test_crop_geometry_by_distance(self):
        sourcepcd=self.pcd
        sourceMesh=self.mesh
        cutterMeshes=[self.slabMesh,self.wallMesh,self.columnMesh,self.beamMesh]
        # mesh + [mesh]
        result1=gt.crop_geometry_by_distance(source=sourceMesh,cutters=cutterMeshes)
        self.assertGreater(len(result1.vertices),1900 )

        # pcd + [mesh]
        result2=gt.crop_geometry_by_distance(source=sourcepcd,cutters=cutterMeshes)
        self.assertGreater(len(result2.points),350000 ) 

        # mesh + pcd
        result3=gt.crop_geometry_by_distance(source=sourceMesh,cutters=sourcepcd)
        self.assertGreater(len(result3.vertices),5000 ) 

        # pcd + mesh
        result4=gt.crop_geometry_by_distance(source=sourcepcd,cutters=sourceMesh)
        self.assertLess(len(result4.points),1200000 ) 
    
    def test_determine_percentage_of_coverage(self):
        reference=self.pcd
        sources=[self.slabMesh,self.wallMesh,self.columnMesh,self.beamMesh]
        percentages=gt.determine_percentage_of_coverage(sources=sources,reference=reference)
        self.assertEqual(len(percentages), len(sources))
        self.assertLess(percentages[0],0.005)
        self.assertLess(percentages[1],0.001)
        self.assertLess(percentages[2],0.4)
        self.assertLess(percentages[3],0.001)

    def test_show_geometries(self):
        # gt.show_geometries((self.bimMeshes + [self.pcd, self.mesh]), color = True)
        self.assertEqual(0,0)
        
    def test_create_3d_camera(self):
        cam=gt.create_3d_camera()
        self.assertIsInstance(cam,o3d.geometry.TriangleMesh)
        
    # def test_save_view_point(self):
    #     print("test_save_view_point")
    #     filename=os.path.join(self.resourcePath,'myView.png')
    #     gt.save_view_point(self.pcd, filename)
    #     self.assertTrue(os.path.exists(filename))

    def test_crop_geometry_by_convex_hull(self):
        sourceMesh=gt.mesh_to_trimesh(self.mesh)
        meshes=[self.slabMesh,self.wallMesh,self.columnMesh,self.beamMesh]
        cutters=[gt.mesh_to_trimesh(mesh) for mesh in meshes]

        innerCrop=gt.crop_geometry_by_convex_hull(source=sourceMesh, cutters=cutters, inside = True )
        self.assertEqual(len(innerCrop),4)
        self.assertGreater(len(innerCrop[0].vertices),12000)
        self.assertGreater(len(innerCrop[1].vertices),800)
        self.assertGreater(len(innerCrop[2].vertices),1000)
        self.assertGreater(len(innerCrop[3].vertices),1000)

        outerCrop=gt.crop_geometry_by_convex_hull(source=sourceMesh, cutters=cutters[0], inside = False ) 
        self.assertGreater(len(outerCrop[0].vertices),255000)         
        
    def test_crop_geometry_by_raycasting(self):
        pcd=self.pcd
        mesh=self.mesh
        hull, _ = pcd.compute_convex_hull()

        innerCrop=gt.crop_geometry_by_raycasting(source=mesh, cutter=hull, inside=True, strict =True ) 
        self.assertLess(len(innerCrop.vertices),200000 )

        innerCrop=gt.crop_geometry_by_raycasting(source=mesh, cutter=hull, inside=True, strict =False ) 
        self.assertLess(len(innerCrop.vertices),200000 )

        outerCrop=gt.crop_geometry_by_raycasting(source=mesh, cutter=hull, inside=False, strict =True ) 
        self.assertLess(len(outerCrop.vertices),200000 )
        
    def test_get_translation(self):
        box=self.mesh.get_oriented_bounding_box()
        centerGt=box.get_center()

        #cartesianBounds
        cartesianBounds=gt.get_cartesian_bounds(box)
        center=gt.get_translation(cartesianBounds)
        self.assertAlmostEqual(math.dist(centerGt,center),0,delta=0.01)

        #orientedBounds
        orientedBounds= np.asarray(box.get_box_points())
        center=gt.get_translation(orientedBounds)
        self.assertAlmostEqual(math.dist(centerGt,center),0,delta=0.01)

        #cartesianTransform
        center=gt.get_translation(self.image2CartesianTransform)
        self.assertAlmostEqual(self.image2CartesianTransform[0][3],center[0],delta=0.01)

    def test_get_mesh_collisions_open3d(self):
        #NOT IMPLEMENTED
        pass
        # meshes=[self.slabMesh,self.wallMesh,self.columnMesh,self.beamMesh]
        # slabCollisions= gt.get_mesh_collisions_open3d(sourceMesh=self.slabMesh, meshes=meshes,t_d=0.5)
        # self.assertEqual(len(slabCollisions),23)

        # wallCollisions= gt.get_mesh_collisions_open3d(sourceMesh=self.wallMesh, meshes=self.bimBoxes)
        # self.assertEqual(len(wallCollisions),1)

        # ColumnCollisions= gt.get_mesh_collisions_open3d(sourceMesh=self.columnMesh, meshes=self.bimBoxes)
        # self.assertEqual(len(ColumnCollisions),23)
       
    def test_get_mesh_collisions_trimesh(self):

        inliers=gt.get_mesh_collisions_trimesh(sourceMesh=self.mesh , geometries =self.bimMeshes) 
        self.assertEqual(len(inliers),66 )

    def test_get_pcd_collisions(self):
        inliers=gt.get_pcd_collisions(sourcePcd=self.pcd, geometries =self.bimMeshes)
        self.assertLess(len(inliers),150 )

    def test_get_rotation_matrix(self):        
        rotationGt=np.array([[ 0.95891633, -0.28368445,  0.00161308],
                            [-0.28365413, -0.95887203, -0.01023566],
                            [ 0.00445043,  0.00935758 ,-0.99994631]])

        #cartesianTransform
        r1=gt.get_rotation_matrix(self.image2CartesianTransform)
        self.assertAlmostEqual(r1[0][0],self.image2CartesianTransform[0][0], delta=0.01)

        #Euler angles
        r2=gt.get_rotation_matrix(np.array([-121.22356551,   83.97341873,   -7.21021157]))
        self.assertAlmostEqual(r2[0][0],self.image2CartesianTransform[0][0], delta=0.01)
        
        #quaternion
        r3=gt.get_rotation_matrix(np.array([ 0.60465535, -0.28690085 , 0.66700836 ,-0.32738304]))
        self.assertAlmostEqual(r3[0][0],self.image2CartesianTransform[0][0], delta=0.01)

        #orientedBounds
        box=self.mesh.get_oriented_bounding_box()
        r4=gt.get_rotation_matrix(np.asarray(box.get_box_points()))
        self.assertAlmostEqual(r4[0][0],rotationGt[0][0], delta=0.01)

    def test_mesh_to_trimesh(self):
        triMesh=gt.mesh_to_trimesh(self.mesh)
        mesh2=triMesh.as_open3d
        self.assertEqual(len(self.mesh.triangles),len(mesh2.triangles))

if __name__ == '__main__':
    unittest.main()



import copy
import math
import os
import shutil
import time
import unittest
from multiprocessing.sharedctypes import Value

import cv2
import geomapi.utils as ut
import ifcopenshell
import numpy as np
import open3d as o3d
import PIL
import pye57
import rdflib
from geomapi.nodes import *
from ifcopenshell.util.selector import Selector
from rdflib import RDF, RDFS, Graph, Literal, URIRef


class TestNode(unittest.TestCase):

################################## SETUP/TEARDOWN CLASS ######################
    @classmethod
    def setUpClass(cls):
        #execute once before all tests
        print('-----------------Setup Class----------------------')
        st = time.time()
        cls.path=os.path.join(os.getcwd(), "test","testfiles") 

        #ONTOLOGIES
        cls.exif = rdflib.Namespace('http://www.w3.org/2003/12/exif/ns#')
        cls.geo=rdflib.Namespace('http://www.opengis.net/ont/geosparql#') #coordinate system information
        cls.gom=rdflib.Namespace('https://w3id.org/gom#') # geometry representations => this is from mathias
        cls.omg=rdflib.Namespace('https://w3id.org/omg#') # geometry relations
        cls.fog=rdflib.Namespace('https://w3id.org/fog#')
        cls.v4d=rdflib.Namespace('https://w3id.org/v4d/core#')
        cls.openlabel=rdflib.Namespace('https://www.asam.net/index.php?eID=dumpFile&t=f&f=3876&token=413e8c85031ae64cc35cf42d0768627514868b2f#')
        cls.e57=rdflib.Namespace('http://libe57.org#')
        cls.xcr=rdflib.Namespace('http://www.w3.org/1999/02/22-rdf-syntax-ns#')
        cls.ifc=rdflib.Namespace('http://ifcowl.openbimstandards.org/IFC2X3_Final#')

        #GRAPH 
        cls.meshGraphPath=os.path.join(cls.path,'meshGraph.ttl')
        cls.meshGraph=Graph().parse(cls.meshGraphPath)
                
        #MESH1
        cls.path1=os.path.join(cls.path,'MESH',"week22.obj")  
        cls.mesh1= o3d.io.read_triangle_mesh(cls.path1)
        cls.subject1=next(s for s in cls.meshGraph.subjects(RDF.type) if "week22" in s.toPython())
        
        #MESH2
        cls.path2=os.path.join(cls.path,"MESH","Basic Wall_211_WA_Ff1_Glued brickwork sandlime 150mm_1095339.obj")  
        cls.mesh2= o3d.io.read_triangle_mesh(cls.path2)
        cls.subject2=next(s for s in cls.meshGraph.subjects(RDF.type) if "Basic_Wall_211_WA_Ff1_Glued_brickwork_sandlime_150mm_1095339" in s.toPython())

        #MESH3
        cls.path3=os.path.join(cls.path,"MESH","Precast Stair_Stair_1181900 Landing 1.obj")  
        cls.mesh3= o3d.io.read_triangle_mesh(cls.path3)
        cls.subject3=next(s for s in cls.meshGraph.subjects(RDF.type) if "Precast_Stair_Stair_1181900_Landing_1" in s.toPython())

        #RESOURCES
        cls.resourcePath=os.path.join(cls.path,"resources")
        if not os.path.exists(cls.resourcePath):
            os.mkdir(cls.resourcePath)
 
        #FILES
        cls.files=ut.get_list_of_files( cls.path)
  
        #TIME TRACKING           
        et = time.time()
        print("startup time: "+str(et - st))
        print('{:50s} {:5s} '.format('tests','time'))
        print('------------------------------------------------------')


    @classmethod
    def tearDownClass(cls):
        #execute once after all tests
        print('-----------------TearDown Class----------------------')
        if os.path.exists(cls.resourcePath):
            shutil.rmtree(cls.resourcePath)  
        if os.path.exists(os.path.join(os.getcwd(),'MESH') ):
            shutil.rmtree(os.path.join(os.getcwd(),'MESH') )
        if os.path.exists(os.path.join(os.getcwd(),'resources') ):
            shutil.rmtree(os.path.join(os.getcwd(),'resources') )

################################## SETUP/TEARDOWN ######################

    def setUp(self):
        #execute before every test
        self.startTime = time.time()   

    def tearDown(self):
        #execute after every test
        t = time.time() - self.startTime
        print('{:50s} {:5s} '.format(self._testMethodName,str(t)))
################################## TEST FUNCTIONS ######################

    def test_meshnode_creation_from_subject(self):
        #subject
        subject='myNode'
        node= MeshNode(subject=subject)
        self.assertEqual(node.subject.toPython(),'file:///'+subject)

        #http://
        subject='http://session_2022_05_20'
        node= MeshNode(subject=subject)
        self.assertEqual(node.subject.toPython(),subject)
        
        #erroneous char       
        subject='[[http://ses>sion_2022_<05_20]]'
        node= MeshNode(subject=subject)
        self.assertEqual(node.subject.toPython(),'http://__ses_sion_2022__05_20__')

    def test_meshnode_creation_from_graph(self):
        subject=next(self.meshGraph.subjects(RDF.type))
        node= MeshNode(graph=self.meshGraph, subject=subject)
        self.assertEqual(node.subject.toPython(),subject.toPython())
        object=self.meshGraph.value(subject,self.v4d['faceCount'])
        self.assertEqual(node.faceCount,object.toPython())
        
    def test_meshnode_creation_from_graph_path(self):
        subject=next(self.meshGraph.subjects(RDF.type))
        node= MeshNode(graphPath=self.meshGraphPath, subject=subject)
        self.assertEqual(node.subject.toPython(),subject.toPython())
        object=self.meshGraph.value(subject,self.v4d['faceCount'])
        self.assertEqual(node.faceCount,object.toPython())
        
    def test_meshnode_creation_from_path(self):
        #path1 without getResource
        node= MeshNode(path=self.path1)
        self.assertEqual(node.name,ut.get_filename(self.path1))
        #path2 with getResource
        node= MeshNode(path=self.path2,getResource=True)        
        self.assertEqual(node.name,ut.get_filename(self.path2))
        self.assertEqual(node.faceCount,len(self.mesh2.triangles))
        #path3 
        node= MeshNode(path=self.path3,getResource=True)
        self.assertEqual(node.name,ut.get_filename(self.path3))
        self.assertEqual(node.faceCount,len(self.mesh3.triangles))

    def test_meshnode_creation_from_resource(self):
        #mesh1
        node= MeshNode(resource=self.mesh1)
        self.assertEqual(node.faceCount,len(self.mesh1.triangles))
        #mesh2
        node= MeshNode(resource=self.mesh2)
        self.assertEqual(node.faceCount,len(self.mesh2.triangles))
        #mesh3
        node= MeshNode(resource=self.mesh3)
        self.assertEqual(node.faceCount,len(self.mesh3.triangles))

    def test_creation_from_subject_and_graph_and_graphPath(self):        
        subject=next(self.meshGraph.subjects(RDF.type))
        node= MeshNode(subject=subject,graph=self.meshGraph,graphPath=self.meshGraphPath)
        self.assertEqual(node.subject.toPython(),subject.toPython())
        node.to_graph()
        self.assertTrue((subject, self.v4d['faceCount'], Literal(node.faceCount)) in self.meshGraph)

    def test_creation_from_subject_and_path(self):        
        node= MeshNode(subject=self.subject2,path=self.path2,getResource=True)
        self.assertEqual(node.subject.toPython(),self.subject2.toPython())
        node.to_graph()
        box= self.mesh2.get_oriented_bounding_box()
        min=box.get_min_bound()
        self.assertAlmostEqual(np.asarray(node.orientedBounds)[0,0],min[0],delta=0.01)

    def test_creation_from_subject_and_path_and_graph(self):        
        node= MeshNode(subject=self.subject3,path=self.path3,graph=self.meshGraph,getResource=True)
        self.assertEqual(node.subject.toPython(),self.subject3.toPython())
        node.to_graph()
        initialGraph=ut.get_subject_graph(self.meshGraph,subject=self.subject3)
        self.assertEqual(len(node.graph),len(initialGraph))
        box= self.mesh3.get_axis_aligned_bounding_box()
        min=box.get_min_bound()
        self.assertAlmostEqual(node.cartesianBounds[0],min[0],delta=0.01)

    def test_creation_from_resource_and_path(self):        
        node= MeshNode(resource=self.mesh1,path=self.path1)
        self.assertEqual(node.subject.toPython(),'file:///'+ut.get_filename(self.path1) )

    def test_creation_from_subject_resource_and_path(self):        
        node= MeshNode(subject=self.subject2,resource=self.mesh2,path=self.path2)
        self.assertEqual(node.subject.toPython(),self.subject2.toPython() )
        
    def test_creation_from_subject_resource_and_path_and_graph(self):        
        node= MeshNode(subject=self.subject3,resource=self.mesh3,path=self.path3, graph=self.meshGraph)
        self.assertEqual(node.subject.toPython(),self.subject3.toPython() )
        node.to_graph()
        object=node.graph.value(node.subject,self.v4d['path'])
        self.assertEqual(object.toPython(),os.path.join("MESH",ut.get_filename(self.path3)+'.obj') )

    def test_node_creation_with_get_resource(self):
        #mesh
        node= MeshNode(resource=self.mesh1)
        self.assertIsNotNone(node.resource)

        #path without getResource
        node= MeshNode(path=self.path2)
        self.assertIsNone(node.resource)

        #path with getResource
        node= MeshNode(path=self.path3,getResource=True)
        self.assertIsNotNone(node.resource)

        #graph with get resource
        node= MeshNode(subject=self.subject2,graph=self.meshGraph,getResource=True)
        self.assertIsNone(node.resource)

        #graphPath with get resource
        node= MeshNode(subject=self.subject3,graphPath=self.meshGraphPath,getResource=True)
        self.assertIsNotNone(node.resource)

    def test_delete_resource(self):
        node= MeshNode(resource=self.mesh1)
        self.assertIsNotNone(node.resource)
        del node.resource
        self.assertIsNone(node.resource)

    def test_save_resource(self):
        #no mesh -> False
        node= MeshNode()
        self.assertFalse(node.save_resource())

        #directory
        node= MeshNode(resource=self.mesh2)
        self.assertTrue(node.save_resource(os.path.join(self.path,'resources')))

        # #graphPath        
        # node= MeshNode(resource=self.mesh2,graphPath=self.meshGraphPath)
        # self.assertTrue(node.save_resource())

        #no path or graphPath
        node= MeshNode(resource=self.mesh2)        
        self.assertTrue(node.save_resource())

        #invalid extension -> error
        node= MeshNode(resource=self.mesh1)
        self.assertRaises(ValueError,node.save_resource,os.path.join(self.path,'resources'),'.kjhgfdfg')

        #.ply -> ok
        node= MeshNode(resource=self.mesh2)
        self.assertTrue(node.save_resource(os.path.join(self.path,'resources'),'.ply'))
        self.assertEqual(node.path,os.path.join(self.path,'resources',node.name+'.ply'))

        #.obj -> ok
        node= MeshNode(resource=self.mesh3)
        self.assertTrue(node.save_resource(os.path.join(self.path,'resources'),'.obj'))
        self.assertEqual(node.path,os.path.join(self.path,'resources',node.name+'.obj'))
        
        #path -> new name
        node= MeshNode(subject=URIRef('myMesh'),path=self.path2,getResource=True)
        self.assertTrue(node.save_resource())
        
        #graphPath with directory
        node=MeshNode(subject=self.subject2,graphPath=self.meshGraphPath, resource=self.mesh3)
        self.assertTrue(node.save_resource(os.path.join(self.path,'resources')))

        #graph with new subject
        node=MeshNode(subject=self.subject3,graph=self.meshGraph, resource=self.mesh3)
        node.name='myMesh'
        self.assertTrue(node.save_resource())

    def test_get_resource(self):
        #mesh
        node=MeshNode(resource=self.mesh3)  
        self.assertIsNotNone(node.get_resource())

        #no mesh
        node=MeshNode()
        self.assertIsNone(node.get_resource())

        #graphPath with getResource
        node=MeshNode(graphPath=self.meshGraphPath,subject=self.subject2,getResource=True)
        self.assertIsNotNone(node.get_resource())

    def test_get_metadata_from_resource(self):
        #mesh
        node=MeshNode(resource=self.mesh3)  
        self.assertIsNotNone(node.orientedBounds)
        self.assertIsNotNone(node.cartesianBounds)
        self.assertIsNotNone(node.cartesianTransform)
        self.assertIsNotNone(node.faceCount)
        self.assertIsNotNone(node.pointCount)

        #graphPath
        node=MeshNode(graphPath=self.meshGraphPath,subject=self.subject2,getResource=True)
        self.assertIsNotNone(node.orientedBounds)
        self.assertIsNotNone(node.cartesianBounds)
        self.assertIsNotNone(node.cartesianTransform)
        self.assertIsNotNone(node.faceCount)
        self.assertIsNotNone(node.pointCount)

if __name__ == '__main__':
    unittest.main()



import unittest
import geomapi.utils.geospatialutils as geo

class TestGeometryutils(unittest.TestCase):

    def test_l72_to_wgs84(self):
        latBel,lngBel=geo.lambert72_to_spherical_coordinates(x =103657.078099938 ,y=194604.105800153)
        LatWGS84,LngWGS84=geo.belgian_datum_to_wgs84(latBel,lngBel)
        # self.assertAlmostEqual(LatWGS84,51.05973284541117,places=3)
        # self.assertAlmostEqual(LngWGS84,3.707712199534464,places=3)

    def test_wgs84_to_l72(self): # this transformation is not correct
        latBel,lngBel=geo.wgs84_to_belgian_datum(LatWGS84=51.05973284541117,LngWGS84=3.707712199534464)
        x,y=geo.spherical_coordinates_to_lambert72(latBel,lngBel)
        # self.assertAlmostEqual(x,103657.078099938,places=3)
        # self.assertAlmostEqual(y,194604.105800153,places=3)
        
    # def test_wgs84_to_l72_GDAL(self):
    #     x,y=geo.wgs84_to_l72(lat=51.05973284541117,long=3.707712199534464) 
    #     self.assertAlmostEqual(x,103657.078099938,places=3)
    #     self.assertAlmostEqual(y,194604.105800153,places=3)

if __name__ == '__main__':
    unittest.main()


"""
BIMNode - a Python Class to govern the data and metadata of BIM data (Open3D, RDF)
"""
#IMPORT PACKAGES
from ast import Raise
import open3d as o3d 
import numpy as np 
from rdflib import Graph, URIRef
import os
import ifcopenshell
import ifcopenshell.geom as geom
import ifcopenshell.util
#IMPORT MODULES
from geomapi.nodes import GeometryNode
import geomapi.utils as ut
import geomapi.utils.geometryutils as gt


class LinesetNode (GeometryNode):
    """
    This class stores a LinesetNode, including the metadata from an ifcElement (OpenShell) and the data from the mesh geometry (Open3D)
    
    Goals:
    - Given a path, import all the linked resources including images, meshes, ect.
    - Convert non-RDF metadata files (.ifc, .obj, ect..) to BIMNodes and export them to RDF
    - use URIRef() to reference the node, ect...
    """
    # class attributes
    # ifcPath = None # (string) path to an ifc file
    # globalId=None # (string) ifc guid of the element e.g. 2DhaSlBmz2puIzqaWbJt0S
    # className=None #(string) ifcElement class e.g. ifcWall
    # mesh = None # (o3d.mesh) mesh geometry of the ifcElement (MB)
    # ifcElement = None # (ifcopenshell.entity_instance)  Don't store an ifcElement in your node. it can't be properly passed around
    # self.path = None # (string) path of the o3d.mesh geometry
    # self.ifc = None # (string) path to an ifc file => you don't want to store the entire ifc file!
    # self.ifcElement = None # (string) 
    # self.pcds = None #(List[URIRef]) => link to other pcd. 
    
    #Classification
    # self.label = None #(string) ??? (MB)
    # self.meshes = None # (List[URIRef]) => link to other meshes. (MB)
    
    #Geometry
    
    # # Monitoring
    # self.phase = None # (string) ??? this can't be really extracted from the IFC (MB)
    # self.progress = None # (double [0.0;1.0]) percentage of completion
    # self.quality = None # ([%LOA30,#LOA20,#LOA10]) inliers Level-of-Accuracy (LOA)
    # self.deviation = None # (np.array(4,4)) cartesian transform 
    # self.cracks = None # (bool) True if cracks are detected on the image texture

    #Questions
    # is phase and timestamp the same?
    # monenclature
    # where do we store the image masks?
    # do we also need a CAD node?

    def __init__(self,  graph : Graph = None, 
                        graphPath: str= None,
                        subject : URIRef = None,
                        path : str= None, 
                        lineset: o3d.geometry.LineSet = None,
                        getResource : bool = False,
                        **kwargs): 
        """Overloaded function.

        Args:
            0.graph (RDFlib Graph) : RDF Graph with a single subject. if no subject is present, the first subject in the Graph will be used to initialise the Node.
            
            1.graphPath (str)  + (optional) subject:  RDF Graph file path with a subject. if no subject is present, the first subject in the Graph will be used to initialise the Node.

            2.path (str) : path to .obj or .ply file (Note that this node will also contain the data)

            3.mesh (o3d.geometry.TriangleMesh) : Open3D TriangleMesh (Note that this node will also contain the data)

            4.ifcPath (str) + globalId (str) + (optional) getResource: (Note that this node will also contain the data)

            5.ifcElement (ifcopenshell.entity_instance) :  Warning, never attach an IfcElement to a node directly as this is very unstable!
                
        Returns:
            A BIMNode with metadata (if you also want the data, call node.get_resource() method)
        """   
        self.path=path
        self.lineset=lineset        
        super().__init__(   graph= graph,
                            graphPath= graphPath,
                            subject= subject,
                            **kwargs)        

        
        if self.lineset is not None: #data + metadata
            if type(lineset) is o3d.geometry.LineSet and len(lineset.vertices) >=2:
                self.get_metadata_from_lineset()
            else:
                raise ValueError('lineset must be o3d.geometry.Lineset and len(lineset.vertices) >=2')
        
        elif path is not None:
            if path.endswith('.obj') or path.endswith('.ply'): #data + metadata
                self.name=ut.get_filename(path)
                self.timestamp=ut.get_timestamp(path)
                self.lineset=o3d.io.read_triangle_mesh(self.path)                
                self.get_metadata_from_lineset()
            else:
                raise ValueError('file must be .obj or .ply')

        if getResource:
            self.get_resource()

    def clear_geometry(self):
        """Clear all geometries in the Node.
        """
        if getattr(self,'mesh',None) is not None:
            self.mesh=None

    def get_mesh(self) -> o3d.geometry.TriangleMesh:
        """Returns the mesh data in the node. If none is present, it will search for the data on drive.

        Returns:
            o3d.geometry.TriangleMesh or None
        """
        if getattr(self,'mesh',None) is not None and len(self.mesh.triangles)>1:
            return self.mesh 
        elif self.get_resource():
            return self.mesh
        return None  

    def get_resource(self)->bool: 
        """
        get o3d.geometry.TriangleMesh from self.path or self.name
        """
        if getattr(self,'mesh',None) is None:
            self.mesh = o3d.io.read_triangle_mesh(self.get_resource_path(fileFormat=ut.MESH_EXTENSION))
                
            if getattr(self,'ifcPath',None) is not None and getattr(self,'globalId',None) is not None :
                try:
                    ifc = ifcopenshell.open(self.ifcPath)   
                    ifcElement= ifc.by_guid(self.globalId)
                    self.mesh=gt.ifc_to_mesh(ifcElement)
                    return True
                except:
                    pass
        elif getattr(self,'mesh',None) is not None:
            if len(self.mesh.triangles)>1:
                return True
        return False
 
    def export_geometry (self, extension = '.ply') ->bool:
        """_summary_

        Args:
            extension (str, optional): file extension. Defaults to '.ply'.

        Raises:
            ValueError: only .ply or .obj are allowed

        Returns:
            bool: return True if export was succesful
        """        
        #check path
        if getattr(self,'mesh',None) is not None:
            if getattr(self,'path',None) is not None and os.path.exists(self.path):
                pass
            else:
                #check name
                if getattr(self,'name',None) is not None:
                    self.name=ut.validate_string(self.name)
                else:
                    self.name=self.guid

                #check session/graphPath and directory
                if getattr(self,'sessionPath',None) is not None :
                    if os.path.exists((self.sessionPath +'\\BIM') ):
                        dir=self.sessionPath +'\\BIM'
                    else:
                        dir=os.mkdir((self.sessionPath+'\\BIM'))
                elif getattr(self,'graphPath',None) is not None: 
                    if os.path.exists(self.graphPath +'\\BIM' ):
                        dir=self.graphPath +'\\BIM'
                    else:
                        dir=os.mkdir((self.graphPath+'\\BIM'))                    
                else:
                    if os.path.exists((os.getcwd()+'\\BIM' )):
                        dir=os.getcwd()+'\\BIM'
                    else:
                        dir=os.mkdir((os.getcwd()+'\\BIM'))
                                    
                #check extension
                if extension == '.ply' or extension == '.obj':
                    self.path=dir+'\\'+self.name+extension
                else:
                    raise ValueError ('only .ply or .obj allowed') 
            try:
                o3d.io.write_triangle_mesh(self.path, self.mesh)
                return True
            except:
                print("Export failed of " + self.name )
        return False

    def get_metadata_from_lineset(self) -> bool:
        self.get_resource()
        if getattr(self,'mesh',None) is not None and len(self.mesh.triangles) >1:
            try:
                center=self.mesh.get_center()  
                self.cartesianTransform= np.array([[1,0,0,center[0]],
                                                    [0,1,0,center[1]],
                                                    [0,0,1,center[2]],
                                                    [0,0,0,1]])
                self.pointCount= len(self.mesh.vertices)
                self.cartesianBounds=gt.get_cartesian_bounds(self.mesh)
            except:
                pass
            return True
        else:
            print('No proper geometries found to extract the metadata. len(self.mesh.triangles) >1')
            return False
    
   


"""
General Basic functions to support other modules
"""

from distutils import extension
from doctest import testfile
from fileinput import filename
import os
import rdflib
from rdflib import XSD, Graph,URIRef,Literal,RDF
import re
import numpy as np
import ntpath
import datetime
import PIL.Image
from PIL.ExifTags import GPSTAGS, TAGS
from typing import List,Tuple
from warnings import warn


#### GLOBAL VARIABLES ####

RDF_EXTENSIONS = [".ttl"]
IMG_EXTENSION = [".jpg", ".png", ".JPG", ".PNG", ".JPEG"]
MESH_EXTENSION = [".obj",".ply",".fbx" ]
PCD_EXTENSION = [".pcd", ".e57",".pts", ".ply", '.xml']
INT_ATTRIBUTES = ['pointCount','faceCount','e57Index'] #'label'
FLOAT_ATTRIBUTES = ['xResolution','yResolution','imageWidth','imageHeight','focalLength35mm','principalPointU','principalPointV','accuracy']
LIST_ATTRIBUTES =  ['distortionCoeficients']

exif = rdflib.Namespace('http://www.w3.org/2003/12/exif/ns#')
geo=rdflib.Namespace('http://www.opengis.net/ont/geosparql#') #coordinate system information
gom=rdflib.Namespace('https://w3id.org/gom#') # geometry representations => this is from mathias
omg=rdflib.Namespace('https://w3id.org/omg#') # geometry relations
fog=rdflib.Namespace('https://w3id.org/fog#')
v4d=rdflib.Namespace('https://w3id.org/v4d/core#')
openlabel=rdflib.Namespace('https://www.asam.net/index.php?eID=dumpFile&t=f&f=3876&token=413e8c85031ae64cc35cf42d0768627514868b2f#')
e57=rdflib.Namespace('http://libe57.org#')
xcr=rdflib.Namespace('http://www.w3.org/1999/02/22-rdf-syntax-ns#')
ifc=rdflib.Namespace('http://ifcowl.openbimstandards.org/IFC2X3_Final#')

#### BASIC OPERATIONS ####

def get_extension(path:str) -> str:
    """Returns the file extension.\n
    E.g. D://myData//test.txt -> .txt

    Args:
        path (str): file path

    Returns:
        extension (str) e.g. '.txt'
    """
    _,extension = os.path.splitext(path)
    return extension

def replace_str_index(text:str,index:int,replacement:str='_'):
    """Replace a string character at the location of the index with the replacement.\n

    Args:
        1. text (str)\n
        2. index (int): index to replace. \n
        3. replacement (str, optional): replacement character. Defaults to '_'.\n

    Returns:
        text (str) with updated characters
    """
    return '%s%s%s'%(text[:index],replacement,text[index+1:])


def get_folder_path(path :str) -> str:
    """Returns the folderPath.

    Args:
        path(str): file path

    Raises:
        FileNotFoundError: If the given graphPath or sessionPath does not lead to a valid systempath

    Returns:
        The path (str) to the folder
    """
    folderPath= os.path.dirname(path)
    if not os.path.isdir(folderPath):
        print("The given path is not a valid folder on this system.")    
    return folderPath

def get_variables_in_class(cls) -> List[str]: 
    """Returns a list of class attributes in the class.\n

    Args:
        cls(class): class e.g. MeshNode
    
    Returns:
        list of class attribute names
    """  
    return [i.strip('_') for i in cls.__dict__.keys() ] 

def get_list_of_files(directoryPath:str) -> list:
    """Get a list of all filepaths in the directory and subdirectories.\n

    Args:
        directoryPath: directory path e.g. \n
            "D:\\Data\\2018-06 Werfopvolging Academiestraat Gent\\week 22\\"
            
    Returns:
        A list of filepaths
    """
    # names in the given directory 
    listOfFile = os.listdir(directoryPath)
    allFiles = list()
    # Iterate over all the entries
    for entry in listOfFile:
        # Create full path
        fullPath = os.path.join(directoryPath, entry)
        # If entry is a directory then get the list of files in this directory 
        if os.path.isdir(fullPath):
            allFiles = allFiles + get_list_of_files(fullPath)
        else:
            allFiles.append(fullPath)                
    return allFiles

def get_subject_graph(graph:Graph, subject:URIRef = None) -> Graph:
        """Returns a subselection of the full Graph that only contains the triples of one subject.\n

        Args:
            1. graph (Graph) \n
            2. subject (URIRef, optional): If no subject is provided, the first one is picked. Defaults to first subject in the graph.
        
        Returns:
            Graph with the triples of one subject
        """
        #input validation       
        if(subject and subject not in graph.subjects()): 
            raise ValueError('subject not in graph')
        elif (not subject): # No subject is defined yet, pick the first one
            subject=next(graph.subjects())        

        #create graph
        newGraph = Graph()
        newGraph += graph.triples((subject, None, None)) 
        newGraph._set_namespace_manager(graph._get_namespace_manager())

        #validate output
        if (len(newGraph) !=0):
            return newGraph
        else:
            return None

def get_paths_in_class(cls) -> list: 
    """Returns list of path attributes in the class.\n

    Args:
        cls (class): class e.g. MeshNode

    Rerturns:
        class atttributes that contain 'path' information e.g. cls.ifcPath, cls.path, etc.
    """  
    from re import search
    return [i.strip('_') for i in cls.__dict__.keys() if search('Path',i) or search('path',i)] 

def get_if_exist(data:dict, key:str):
    """Return data of a certain key in a dictionary:

    Args:
        1. data (dict): dictionay with data \n
        2. key (str): keyword attribute in dictionary e.g. data['myattribute'] \n

    Returns:
        data (python values)
    """
    if key in data:
        return data[key]
    return None

def filter_exif_gps_data(data, reference:str) -> float:
    """Returns decimal degrees from world angles in exif data.\n

    Args:
        data (float or tuple): decimal degrees or (degrees, minutes, seconds) notation of world angles\n
        reference (str): direction character i.e. 'S','W','s' or 'w'.\n

    Returns:
        decimal degrees (float)
    """
    if type(data) is float:
        if reference in ['S','W','s','w']:
            return data*-1
        else:
            return data
    elif type(data) is tuple and len(data) == 3:
        value=dms2dd(data[0],data[1],data[2],reference)
        return value     

def dms2dd(degrees:float, minutes:float, seconds:float, direction:str) -> float:
    """Convert world angles (degrees, minutes, seconds, direction) to decimal degrees.\n

    Args:
        1. degrees (float) \n
        2. minutes (float) \n
        3. seconds (float) \n
        4. direction (str): 'N' and 'E' are positive, 'W' and 'S' are negative  \n

    Returns:
        decimal degrees (float)
    """
    dd = float(degrees) + float(minutes)/60 + float(seconds)/(60*60)
    if direction == 'W' or direction == 'S':
        dd *= -1
    return dd

def dd2dms(deg) -> List[float]:
    """Convert decimal degrees to List[degrees, minutes, seconds] notation. \n

    **NOTE**: Direction is not determined (should be easy based on sign).

    Args:
        decimal degrees (float)

    Returns:
        List[degrees(float),minutes(float),seconds(float)]
    """
    d = int(deg)
    md = abs(deg - d) * 60
    m = int(md)
    sd = (md - m) * 60
    return [d, m, sd]

def parse_dms(dms) :
    """Returns decimal degrees (float) notation of geospatial coordinates from various (degrees, minutes, seconds, direction) notations.\n

    Args (degrees, minutes, seconds,direction):
        1. dms (np.array): \n
        2. dms (tuple[float,float,float,float])\n
        3. dms (str): stringed list with values\n

    Raises:
        ValueError: 'dms.size!=4'

    Returns:
        decimal degrees (float)
    """
    try:
        if type(dms) is np.ndarray and dms.size==4:
            return dms2dd(dms[0],dms[1],dms[2], dms[3] )
        elif type(dms) is tuple and len(dms)==4:
            return dms2dd(dms[0],dms[1],dms[2], dms[3] )
        elif type(dms) is str and 'None' not in dms:
            temp=validate_string(dms, ' ')
            temp=temp.replace("\n","")
            temp=temp.replace("\r","")
            temp=temp.split(' ')
            temp=[x for x in temp if x]
            if temp:
                res=np.asarray(temp) 
                return dms2dd(res[0],res[1],res[2], res[3] )
        return None  
    except:
        raise ValueError ('dms.size!=4')

def get_exif_data(img:PIL.Image)-> dict:
    """Returns a dictionary from the exif data of an Image (PIL) item. Also
    converts the GPS Tags.\n

    https://pillow.readthedocs.io/en/stable/reference/Image.html
    
    Args:
        Img (PIL.image)
    
    Returns:
        exifData (dict): dictionary containing all meta data of the PIL Image.
    """
    exifData = {}
    info = img._getexif()
    if info:
        for tag, value in info.items():
            decoded = TAGS.get(tag, tag)
            if decoded == "GPSInfo":
                gps_data = {}
                for t in value:
                    sub_decoded = GPSTAGS.get(t, t)
                    gps_data[sub_decoded] = value[t]

                exifData[decoded] = gps_data
            else:
                exifData[decoded] = value        
        return exifData      
    else:
        return None        

def get_filename(path :str) -> str:
    """ Deconstruct filepath and return filename.
    
    Args:
        path (str): filepath
        
    Returns:
        filename (str)
    """   
    path=ntpath.basename(path)
    _, tail = ntpath.split(path)
    array=tail.split('.')
    return array[0]

def get_folder(path :str) -> str:
    """ Deconstruct path and return forlder
    
    Args:
        path (str): filepath
        
    Returns:
        folderPath (str)
    """
    return os.path.dirname(os.path.realpath(path))

def get_timestamp(path : str) -> str:
    """Returns the timestamp ('%Y-%m-%dT%H:%M:%S') from a filepath.

    Args:
        path (str): filepath

    Returns:
        dateTime (str): '%Y-%m-%dT%H:%M:%S'
    """
    ctime=os.path.getctime(path)
    dtime=datetime.datetime.fromtimestamp(ctime).strftime('%Y-%m-%dT%H:%M:%S')
    return dtime

#### CONVERSIONS ####

def literal_to_cartesianTransform(literal:Literal) -> np.array:
    """Returns cartesianTransform from rdflib.literal. This function is used to convert a serialized (str) to a usable np.array. 

    Args:
        literal (URIRef):  stringed list or np array of the cartesianTransform e.g. \n
            e57:cartesianTransform ""[[ 3.48110207e-01  9.37407536e-01  9.29487057e-03  2.67476305e+01]
                                    [-9.37341584e-01  3.48204779e-01 -1.20077869e-02  6.17326932e+01]
                                    [-1.44927083e-02 -4.53243552e-03  9.99884703e-01  4.84636987e+00]
                                    [ 0.00000000e+00  0.00000000e+00  0.00000000e+00  1.00000000e+00]]""

    Returns:
        cartesianTransform (np.array([4x4])): geometric transform of a Node/Resource.   
    """   
    temp=str(literal)
    try:
        if 'None' not in temp:
            temp=validate_string(temp, ' ')
            temp=temp.replace("\n","")
            temp=temp.replace("\r","")
            temp=temp.split(' ')
            temp=[x for x in temp if x]
            if temp:
                res = list(map(float, temp))   
                res=np.reshape(res,(4,4))
                return np.asarray(res)  
        return None  
    except:
        raise ValueError

def literal_to_array(literal: Literal) -> np.array:
    """Returns array from rdflib.literal.\n

    Args:
        literal (rdflib): literal containing serialized list or array 

    Raises:
        ValueError: 'Conversion error'

    Returns:
        np.array 
    """
    temp=str(literal)
    return np.asarray(string_to_list(temp))

def literal_to_orientedBounds(literal: Literal)-> np.array:
    """Returns orientedBounds from rdflib.literal. This function is used to convert a serialized (str) to a usable np.array. 

    Args:
        literal (URIRef):  stringed list or np array of the orientedBounds e.g. \n
            v4d:orientedBounds  ""[[-1.96025758e+01  1.65884155e+02  2.22874746e+01]\n
                                [ 1.22465470e+02  1.23859440e+02  2.29468276e+01]\n
                                [-5.26111779e+01  5.43129133e+01  2.33762930e+01]\n
                                [-1.95654774e+01  1.65648750e+02 -7.09825603e-01]\n
                                [ 8.94939663e+01  1.20527928e+01  1.03834566e+00]\n
                                [-5.25740795e+01  5.40775081e+01  3.78992731e-01]\n
                                [ 1.22502568e+02  1.23624035e+02 -5.04726756e-02]\n
                                [ 8.94568679e+01  1.22881979e+01  2.40356459e+01]]""

    Returns:
        orientedBounds (np.array([8x3])): 8 bounding coordinates of a geometric asset.   
    """   
    temp=str(literal)
    try:
        if 'None' not in temp:
            temp=validate_string(temp, ' ')
            temp=temp.replace("\n","")
            temp=temp.replace("\r","")
            temp=temp.split(' ')
            temp=[x for x in temp if x]
            if temp:
                res = list(map(float, temp))   
                res=np.reshape(res,(8,3))
                return np.asarray(res)  
        return None  
    except:
        raise ValueError

def literal_to_float(literal: Literal) -> float:
    """Returns float from rdflib.literal 

    Args:
        literal (rdflib): literal containing value

    Raises:
        ValueError: 'Conversion error'

    Returns:
        float 
    """
    try:
        if 'None' in literal:
            return None
        return float(literal.toPython())
    except:
        raise ValueError('Conversion error')

def literal_to_string(literal: Literal)->str:
    """Returns string from rdflib.literal.\n

    Args:
        literal (rdflib): literal containing value

    Returns:
        string
    """
    string=str(literal)
    try:
        if 'None' in string:
            return None
        else:
            return string
    except:
        raise ValueError

def literal_to_list(literal: Literal)->list:
    """Returns list from rdflib.literal.

    Args:
        literal (rdflib): literal containing value

    Raises:
        ValueError: 'Conversion error'

    Returns:
        int 
    """
    string=str(literal)
    try:
        if 'None' in string:
            return None
        else: 
            return string_to_list(string)
    except:
        raise ValueError

def literal_to_int(literal: Literal) -> int:
    """Returns int from rdflib.literal.

    Args:
        literal (rdflib): literal containing value

    Raises:
        ValueError: 'Conversion error'

    Returns:
        int 
    """
    try:
        if 'None' in literal:
            return None
        return int(literal.toPython())
    except:
        raise ValueError ('Conversion error')

def literal_to_uriref(literal: Literal)->URIRef:
    """Returns rdflib.URIRef from rdflib.literal.

    Args:
        literal (rdflib): literal containing value

    Raises:
        ValueError: 'float causes errors'

    Returns:
        URIRef: 
    """
    try:
        temp=literal.toPython()
        if type(temp) is float or type(temp) is int:
            raise ValueError('float causes errors')
        elif 'None' not in literal:            
            return URIRef(literal.toPython())      
        return None
    except:
        raise ValueError

def check_if_subject_is_in_graph(graph:Graph,subject:URIRef) ->bool:
    """Returns True if a subject is present in the Graph. 

    Args:
        graph (Graph): Graph to parse.\n
        subject (URIRef): subject to search. The function only uses the main body of the subject so no prefix mistakes can be made.\n

    Returns:
        bool: True if subject is present.
    """
    testSubject=subject.split('/')[-1]
    for s in graph.subjects():
        graphSubject= s.split('/')[-1]
        if testSubject==graphSubject:
            return True
    return False

def get_graph_intersection(graphs:List[Graph]) -> Graph:
    """Returns the intersection of multiple graphs i.e. all triples of common subjects.\n

    Args:
        graphs (List[Graph]):

    Returns:
        Graph of the intersection
    """
    #retrieve common subjects
    intersectionGraph=graphs[0]
    for i in range(len(graphs)-1):
        intersectionGraph=intersectionGraph & graphs[i+1]
    subjects=[s for s in intersectionGraph.subjects(RDF.type)]
    
    # if no overlap, return None
    if len(subjects)==0:
        return None

    #merge graphs
    joinedGraph=Graph()
    joinedGraph=bind_ontologies(joinedGraph)
    for graph in graphs:
        joinedGraph+=graph

    #select all relevant graphs
    selectGraph=Graph()
    selectGraph=bind_ontologies(selectGraph)
    for s in subjects:
        selectGraph+=get_subject_graph(joinedGraph,s)

    return selectGraph

def get_graph_subject(graph:Graph,subject:URIRef) ->URIRef:
    """Returns graph subject with its appropriate prefix given the main body of a subject. This function is mainly used for local file serializations that mud the graph subjects. You can then use this function to use the correct graph subject regardless of the prefix.\n

    Args:
        1. graph (Graph): graph with subjects with prefix e.g. URIRef(D:\\DATA\\Subject1)\n
        2. subject (URIRef): subject to search for. e.g. URIREF(Subject1)\n

    Raises:
        ValueError: 'subject not in graph.'

    Returns:
        URIRef(subject): e.g.  URIRef(D:\\DATA\\Subject1)
    """
    testSubject=subject.split('/')[-1]
    for s in graph.subjects():
        graphSubject= s.split('/')[-1]
        if testSubject==graphSubject:
            return s
    raise ValueError ('subject not in graph.')

def literal_to_linked_subjects(string:str) -> list:
    """Returns list of URI subjects that were serialized as a string.\n

    Args:
        string (str): e.g. '[ file:///Subject1, http://Subject2]'

    Returns:
        list[items] e.g. [ file:///Subject1, http://Subject2]
    """
    temp=string.split(',')
    temp=[re.sub(r"[\[ \' \]]",'',s) for s in temp]
    return temp

def string_to_list(string:str)->list:
    """Convert string of items to a list of their respective types. Function deals with both np.array and list encodings of the stringed values.\n

    Args:
        string (str): string to parse. 

    Returns:
        List[items in string]
    """
    try:
        if 'None' not in string:
            temp=validate_string(string, ' ')
            temp=temp.replace("\n","")
            temp=temp.replace("\r","")
            temp=temp.split(' ')
            temp=[x for x in temp if x]
            # res = list(map(float, temp))  
            if temp:
                res=[]
                for item in temp:      
                    if is_float(item): 
                        res.append(float(item))
                    elif is_int(item): 
                        res.append(int(item))
                    elif is_string(item): 
                        res.append(str(item))
                    elif is_uriref(item): 
                        res.append(URIRef(item)) 
                return res
        return None  
    except:
        raise ValueError

def string_to_rotation_matrix(string :str) -> np.array:
    """Returns rotation matrix (np.array(3x3)) from string.

    Args:
        string (str): string cast list or np.array that contains the values of the rotation matrix.\n

    Raises:
        ValueError: 'array.size!=9'

    Returns:
        np.array (3x3)
    """
    array=np.asarray(string_to_list(string))
    if array.size==9:
        return np.reshape(array,(3,3))
    else:
        raise ValueError('array.size!=9')

def xml_to_float(xml) -> float:
    """Cast XML string value to float if possible.

    Args:
        xml value

    Returns:
        float of value
    """
    if xml is None:
        return None
    else:
        return float(xml)

def xcr_to_lat(xcr:str) -> float:
    """Returns latitude from XCR serialization. This includes interpretation 'N' and 'S' geospatial values.

    Args:
        xcr (str)

    Returns:
        float of value
    """
    if 'None' in xcr:
        return None
    else:
        list=list=re.findall(r'[A-Za-z]+|\d+(?:\.\d+)?', xcr)
        if 'N' in list[-1]:
            return float(list[0])
        elif 'S' in list[-1]:
            return - float(list[0])

def xcr_to_long(xcr:str) -> float:
    """Returns longitude from XCR serialization. This includes interpretation 'E' and 'W' geospatial values.

    Args:
        xcr (str): 

    Returns:
        float of value
    """
    if 'None' in xcr:
        return None
    else:        
        list=list=re.findall(r'[A-Za-z]+|\d+(?:\.\d+)?', xcr)
        if 'E' in list[-1]:
            return float(list[0])
        elif 'W' in list[-1]:
            return - float(list[0])

def xcr_to_alt(xcr:str) -> float:
    """Returns altitude from XCR serialized height value. This value is sometimes encoded as a fracture 10000/1600.\n

    Args:
        xcr (str): value to process

    Returns:
        float of value
    """
    if 'None' in xcr:
        return None
    else:
        list=list=re.findall(r'[A-Za-z]+|\d+(?:\.\d+)?', xcr)
        if list:
            return float(list[0])/float(list[-1])       

def cartesianTransform_to_literal(matrix : np.array) -> str:
    """ convert nparray [4x4] to str([16x1]).

    Args:
        nparray [4x4]

    Returns:
        List[16]
    """
    if matrix.size == 16: 
        return str(matrix.reshape(16,1))
    else:
        Exception("wrong array size!")    

def featured3d_to_literal(value) -> str:
    "No feature implementation yet"

def featured2d_to_literal(value) -> str:
    "No feature implementation yet"

def item_to_list(item)-> list:
    """Returns [item] if item is not yet a list. This function protects functions that rely on list functionality. 

    Args:
        item (Python value) 

    Returns:
        list[item]
    """
    if type(item) is np.ndarray:
        item=item.flatten()
        return item.tolist()
    elif type(item) is np.array:
        item=item.flatten()
        return item.tolist()
    elif type(item) is list:
        return item
    else:
        return [item]

#### VALIDATION ####

def check_if_uri_exists(list:List[URIRef], subject:URIRef) ->bool:
    """Returns True is a subject occurs in a list of URIRefs.

    Args:
        list (List[URIRef]): reference list
        subject (URIRef): subject to test

    Returns:
        bool: True if subject occurs in list
    """
    list=item_to_list(list)
    list=[item.toPython() for item in list]
    subject=subject.toPython()
      
    if any(subject in s for s in list):
        return True
    else: 
        return False

def get_subject_name(subject:URIRef) -> str:
    """Get the main body of a URIRef graph subject

    Args:
        subject (URIRef)

    Returns:
        str
    """
    string=subject.toPython()
    return string.split('/')[-1]   

def validate_string(string:str, replacement ='_') -> str:
    """Checks path validity. A string is considered invalid if it cannot be serialized by rdflib or is not Windows subscribable.\n
    If not valid, The function adjusts path naming to be Windows compatible.\n

    Features (invalid characters):
        "()^[^<>{}[] ~`],|*$ /\:"\n

    Args:
        path (str): string to check
        replacement (character): characters to replace invalid charters in the string 

    Returns:
        str: cleansed string
    """
    prefix=''
    if 'file:///' in string:
        string=string.replace('file:///','')
        prefix='file:///'
    elif 'http://' in string:
        string=string.replace('http://','')
        prefix='http://'
    for idx,element in enumerate(string):
        if element in "()^[^<>{}[] ~`],|*$ /\:": #
            string=replace_str_index(string,index=idx,replacement=replacement)
    string=prefix+string
    return string

def is_float(element) -> bool:
    """Returns True if Python value is a float.

    Args:
        element (Python value)

    Returns:
        bool: True if Python value is a float
    """
    try:
        float(element)
        return True
    except ValueError:
        return False

def is_int(element) -> bool:
    """Returns True if Python value is an int.

    Args:
        element (Python value)

    Returns:
        bool: True if Python value is an int
    """
    try:
        int(element)
        return True
    except ValueError:
        return False

def is_string(element) -> bool:
    """Returns True if Python value is a str.

    Args:
        element (Python value)

    Returns:
        bool: True if Python value is a str
    """
    #special_characters = "[!@#$%^&*()-+?_= ,<>/]'"
    try:
        str(element)
        return True
    except ValueError:
        return False

def is_uriref(element) -> bool:
    """Returns True if Python value is a rdflib.URIRef.

    Args:
        element (Python value)

    Returns:
        bool: True if Python value is a rdflib.URIRef
    """
    try:
        if type(element) is float:
            return False
        else:
            URIRef(element)
            return True
    except ValueError:
        return False

# #### RDF ####

def get_attribute_from_predicate(graph: Graph, predicate : Literal) -> str:
    """Returns the attribute witout the namespace.

    Args:
        graph (Graph): The Graph containing the namespaces
        predicate (Literal): The Literal to convert

    Returns:
        str: The attribute name
    """
    predStr = str(predicate)
    #Get all the namespaces in the graph
    for nameSpace in graph.namespaces():
        nameSpaceStr = str(nameSpace[1])
        if(predStr.__contains__(nameSpaceStr)):
            predStr = predStr.replace(nameSpaceStr, '')
            break
    return predStr

def bind_ontologies(graph : Graph) -> Graph:
    """Returns a graph that binds in its namespace the ontologies that GEMOMAPI uses and that are not in the rdflib.\n\n

     Features (ontologies):
        1. exif = rdflib.Namespace('http://www.w3.org/2003/12/exif/ns#') \n
        2. geo=rdflib.Namespace('http://www.opengis.net/ont/geosparql#') \n
        3. gom=rdflib.Namespace('https://w3id.org/gom#') \n
        4.  omg=rdflib.Namespace('https://w3id.org/omg#') \n
        5. fog=rdflib.Namespace('https://w3id.org/fog#')\n
        6. v4d=rdflib.Namespace('https://w3id.org/v4d/core#')\n
        7. openlabel=rdflib.Namespace('https://www.asam.net/index.php?eID=dumpFile&t=f&f=3876&token=413e8c85031ae64cc35cf42d0768627514868b2f#')\n
        8. e57=rdflib.Namespace('http://libe57.org#')\n
        9. xcr=rdflib.Namespace('http://www.w3.org/1999/02/22-rdf-syntax-ns#')\n
        10. ifc=rdflib.Namespace('http://ifcowl.openbimstandards.org/IFC2X3_Final#')\n

    Returns:
        Graph with update namespace 
    """
    exif = rdflib.Namespace('http://www.w3.org/2003/12/exif/ns#')
    graph.bind('exif', exif)
    geo=rdflib.Namespace('http://www.opengis.net/ont/geosparql#') #coordinate system information
    graph.bind('geo', geo)
    gom=rdflib.Namespace('https://w3id.org/gom#') # geometry representations => this is from mathias
    graph.bind('gom', gom)
    omg=rdflib.Namespace('https://w3id.org/omg#') # geometry relations
    graph.bind('omg', omg)
    fog=rdflib.Namespace('https://w3id.org/fog#')
    graph.bind('fog', fog)
    v4d=rdflib.Namespace('https://w3id.org/v4d/core#')
    graph.bind('v4d', v4d)
    v4d3D=rdflib.Namespace('https://w3id.org/v4d/3D#')
    graph.bind('v4d3D', v4d3D)
    openlabel=rdflib.Namespace('https://www.asam.net/index.php?eID=dumpFile&t=f&f=3876&token=413e8c85031ae64cc35cf42d0768627514868b2f#')
    graph.bind('openlabel', openlabel)
    e57=rdflib.Namespace('http://libe57.org#')
    graph.bind('e57', e57)
    xcr=rdflib.Namespace('http://www.w3.org/1999/02/22-rdf-syntax-ns#')
    graph.bind('xcr', xcr)
    ifc=rdflib.Namespace('http://ifcowl.openbimstandards.org/IFC2X3_Final#')
    graph.bind('ifc', ifc)
    return graph

def clean_attributes_list(list:list) -> list:
    """Returns a 'cleaned' list of class attributes that are not to be serialized in the graph. This includes data attributes such as resources (point clouds, images), linkedNodes, etc.\n

    **NOTE**: RDF Graphs are meant to store metadata, not actual GBs of data that should be contained in their native file formats.\n

    Args:
        list (list): list of class attributes

    Returns:
        list (list): 'cleaned' list of class attributes 
    """
    #NODE
    excludedList=['graph','resource','graphPath','subject','resource','fullResourcePath','kwargs', 'orientedBoundingBox','type']
    #BIMNODE    
    excludedList.extend(['ifcElement'])
    #MESHNODE
    excludedList.extend(['mesh'])
    #IMGNODE
    excludedList.extend(['exifData','xmlData','image','features2d','pinholeCamera'])
    #PCDNODE
    excludedList.extend(['pcd','e57Pointcloud','e57xmlNode','e57image','features3d'])
    #SESSIONNODE
    excludedList.extend(['linkedNodes'])

    cleanedList = [ elem for elem in list if elem not in excludedList]
    return cleanedList

def check_if_path_is_valid(path:str)-> bool:
    """Returns True if the given path is an existing folder.\n

    Args:
        path (str): path to folder or file

    Returns:
        bool: True if exsists.
    """
    folder=get_folder(path)
    if os.path.isdir(path):
        return True
    elif os.path.exists(folder):
        return True
    else:
        return False

def validate_timestamp(value) -> datetime:
    """Format value as datetime ("%Y-%m-%dT%H:%M:%S")

    Args:
        1.value ('Tue Dec  7 09:38:13 2021')\n
        2.value ("1648468136.033126")\n
        3.value ("2022:03:13 13:55:30")\n
        4.value (datetime)\n
        5.value ("2022-03-13 13:55:30")\n

    Raises:
        ValueError: 'timestamp formatting ("%Y-%m-%dT%H:%M:%S") failed for tuple, datetime and string formats.'

    Returns:
        datetime
    """
    string=str(value)
  
    try:
        return datetime.datetime.strptime(string, "%Y-%m-%dT%H:%M:%S").strftime("%Y-%m-%dT%H:%M:%S") 
    except:
        pass
    try:         
        test=datetime.datetime.strptime(string, "%Y:%m:%d %H:%M:%S").strftime("%Y-%m-%dT%H:%M:%S") 
        return test
    except:
        pass
    try:         
        test=datetime.datetime.strptime(string, "%Y-%m-%d %H:%M:%S").strftime("%Y-%m-%dT%H:%M:%S") 
        return test
    except:
        pass
    try:         
        test=datetime.datetime.strptime(string, "%a %b %d %H:%M:%S %Y").strftime("%Y-%m-%dT%H:%M:%S") 
        return test
    except:
        pass
    try:
        return datetime.datetime.fromtimestamp(float(string)).strftime('%Y-%m-%dT%H:%M:%S')
    except:
        raise ValueError('no valid time formatting found e.g. 1.value (Tue Dec  7 09:38:13 2021) 2.value (1648468136.033126) 3.value (2022:03:13 13:55:30)')

def get_node_resource_extensions(objectType:str) -> list:
    """Returns the potential file formats for different types of Nodes' resources (images, pcd, ortho, etc.).\n

    Args:
        objectType (str): Type of node class

    Returns:
        list with possible extensions e.g. .obj, .ply for MeshNodes
    """    
    if 'MeshNode' in objectType:        
        return MESH_EXTENSION
    if 'SessionNode' in objectType:        
        return MESH_EXTENSION
    elif 'BIMNode' in objectType:        
        return MESH_EXTENSION
    elif 'PointCloudNode' in objectType:        
        return PCD_EXTENSION    
    elif 'ImageNode' in objectType:        
        return IMG_EXTENSION
    elif 'OrthoNode' in objectType:        
        return IMG_EXTENSION
    else:
        return ['.txt']+MESH_EXTENSION+PCD_EXTENSION+IMG_EXTENSION+RDF_EXTENSIONS

def get_node_type(objectType:str) -> URIRef:
    """Return the type of Node as an rdflib literal. By default, 'Node' is returned.

    Features:
        1. MeshNode\n
        2. BIMNode\n
        3. PointCloudNode\n
        4. GeometryNode\n
        5. ImageNode\n
        6. OrthoNode\n
        7. SessionNode\n
        8. linesetNode\n
        9. Node\n

    Args:
        objectType (str): Type of node class.\n

    Returns:
        URIRef (nodeType)
    """
    if 'MeshNode' in objectType:        
        return v4d['MeshNode']
    elif 'BIMNode' in objectType:        
        return v4d['BIMNode']
    elif 'PointCloudNode' in objectType:        
        return v4d['PointCloudNode']
    elif 'GeometryNode' in objectType:        
        return v4d['GeometryNode']
    elif 'ImageNode' in objectType:        
        return v4d['ImageNode']
    elif 'OrthoNode' in objectType:        
        return v4d['OrthoNode']
    elif 'SessionNode' in objectType:        
        return v4d['SessionNode']
    elif 'linesetNode' in objectType:        
        return v4d['linesetNode']
    else:
        return v4d['Node']

def get_data_type(value) -> XSD.ENTITY:
    """Return XSD dataType of Python value. By default, string is returned as the XSD.entity.\n

    Args:
        value (any): data of any Python value (boolean, int, float, dateTime, string)

    Returns:
        XSD.ENTITY 
    """
    if 'bool' in str(type(value)):        
        return XSD.boolean
    elif 'int' in str(type(value)):  
        return XSD.integer
    elif 'float' in str(type(value)):  
        return XSD.float
    elif 'date' in str(type(value)):  
        return XSD.dateTime   
    else:
        return XSD.string

def match_uri(attribute :str) -> URIRef:
    """ Returns fitting predicate from class attribute given the following prefix ontologies.\n
    By default, the v4d ontology[attributeName] is returned.

    Features (ontologies):
        1. exif = rdflib.Namespace('http://www.w3.org/2003/12/exif/ns#') \n
        2. geo=rdflib.Namespace('http://www.opengis.net/ont/geosparql#') \n
        3. gom=rdflib.Namespace('https://w3id.org/gom#') \n
        4.  omg=rdflib.Namespace('https://w3id.org/omg#') \n
        5. fog=rdflib.Namespace('https://w3id.org/fog#')\n
        6. v4d=rdflib.Namespace('https://w3id.org/v4d/core#')\n
        7. openlabel=rdflib.Namespace('https://www.asam.net/index.php?eID=dumpFile&t=f&f=3876&token=413e8c85031ae64cc35cf42d0768627514868b2f#')\n
        8. e57=rdflib.Namespace('http://libe57.org#')\n
        9. xcr=rdflib.Namespace('http://www.w3.org/1999/02/22-rdf-syntax-ns#')\n
        10. ifc=rdflib.Namespace('http://ifcowl.openbimstandards.org/IFC2X3_Final#')\n

    Returns:
        URIRef: predicate 
    """
    #OPENLABEL
    if attribute in ['timestamp','sensor']:
        return openlabel[attribute]
    #E57
    elif attribute in ['cartesianBounds','cartesianTransform','geospatialTransform','pointCount','e57XmlPath','e57Path','e57Index','e57Image']:
        return  e57[attribute]
    #GOM
    elif attribute in ['coordinateSystem']:
        return  gom[attribute]
    #IFC
    elif attribute in ['ifcPath','className','globalId','phase','ifcName']:
        return  ifc[attribute]
    #EXIF
    elif attribute in ['xResolution','yResolution','resolutionUnit','imageWidth','imageHeight']:
        return  exif[attribute]
    #XCR
    elif attribute in ['focalLength35mm','principalPointU','principalPointV','distortionCoeficients','gsd']:
        return  xcr[attribute]
    #XCR
    elif attribute in ['isDerivedFromGeometry']:
        return  omg[attribute]
    #V4D
    else:
        return v4d[attribute]



"""
MeshNode - a Python Class to govern the data and metadata of mesh data (Open3D, Trimesh). \n

This node builds upon the Open3D and Trimesh API for the point cloud definitions.\n
It inherits from GeometryNode which in turn inherits from Node.\n
Be sure to check the properties defined in those abstract classes to initialise the Node.

"""
#IMPORT PACKAGES
from multiprocessing.sharedctypes import Value
import open3d as o3d 
import numpy as np 
from rdflib import Graph, URIRef
import os
import trimesh

#IMPORT MODULES
from geomapi.nodes import GeometryNode
import geomapi.utils as ut
import geomapi.utils.geometryutils as gt

class MeshNode (GeometryNode):   
    def __init__(self,  graph : Graph = None, 
                        graphPath:str=None,
                        subject : URIRef = None,
                        path : str=None, 
                        getResource : bool = False,
                        getMetaData : bool = True,
                        **kwargs): 
        """
        Creates a MeshNode. Overloaded function.\n
        This Node can be initialised from one or more of the inputs below.\n
        By default, no data is imported in the Node to speed up processing.\n
        If you also want the data, call node.get_resource() or set getResource() to True.\n

        Args:\n
            0.graph (RDFlib Graph) : Graph with a single subject (if multiple subjects are present, only the first will be used to initialise the MeshNode)\n
            
            1.graphPath (str):  Graph file path with a single subject (if multiple subjects are present, only the first will be used to initialise the MeshNode)\n

            2.path (str) : path to mesh file (Note that this node will also contain the data)\n

            3.resource (o3d.geometry.TriangleMesh) : Open3D point cloud file (Note that this node will also contain the data)\n
                
            getResource (bool, optional= False): If True, the node will search for its physical resource on drive \n
            getMetaData (bool, optional= True): If True, the node will attempt to extract geometric metadata from the resource if present (cartesianBounds, etc.) \n
        
        Returns:
            A MeshNode with metadata
        """   
        #private attributes
        self.pointCount = None 
        self.faceCount = None        

        super().__init__(   graph= graph,
                            graphPath= graphPath,
                            subject= subject,
                            path=path,
                            **kwargs)        

        #initialisation functionality
        if getResource:
            self.get_resource()
        
        if getMetaData:
            self.get_metadata_from_resource()

#---------------------METHODS----------------------------

    def set_resource(self, value): 
        """Set the self.resource (o3d.geometry.TriangleMesh) of the Node.\n

        Args:
            1. open3d.geometry.TriangleMesh
            2. trimesh.Trimesh

        Raises:
            ValueError: Resource must be an o3d.geometry.TriangleMesh with len(resource.triangles) >=2 or an trimesh.Trimesh instance.
        """
        if 'TriangleMesh' in str(type(value)) and len(value.triangles) >=2:
            self._resource = value
        elif 'Trimesh' in str(type(value)):
            self._resource=value.as_open3d
        else:
            raise ValueError('Resource must be an o3d.geometry.TriangleMesh with len(resource.triangles) >=2 or an trimesh.Trimesh instance.')

    def get_resource(self)->o3d.geometry.TriangleMesh: 
        """Returns the mesh data in the node. \n
        If none is present, it will search for the data on drive from path, graphPath, name or subject. 

        Returns:
            o3d.geometry.TriangleMesh or None
        """
        if self.resource is not None and len(self.resource.triangles)>=2:
            pass
        elif self.get_path():
            resource =  o3d.io.read_triangle_mesh(self.path)
            if len(resource.triangles)>2:
                self._resource  =resource
        return self._resource  

    def save_resource(self, directory:str=None,extension :str = '.ply') ->bool:
        """Export the resource of the Node.\n

        Args:
            directory (str, optional): directory folder to store the data.\n
            extension (str, optional): file extension. Defaults to '.ply'.\n

        Raises:
            ValueError: Unsuitable extension. Please check permitted extension types in utils._init_.\n

        Returns:
            bool: return True if export was succesful
        """         
        #check path
        if self.resource is None:
            return False
        
        #validate extension
        if extension not in ut.MESH_EXTENSION:
            raise ValueError('Invalid extension')

        # check if already exists
        if directory and os.path.exists(os.path.join(directory,self.name + extension)):
            self.path=os.path.join(directory,self.get_name() + extension)
            return True
        elif not directory and self.get_path() and os.path.exists(self.path) and extension in ut.MESH_EXTENSION:
            return True
                    
        #get directory
        if (directory):
            pass    
        elif self.path is not None:    
            directory=ut.get_folder(self.path)            
        elif(self.graphPath): 
            dir=ut.get_folder(self.graphPath)
            directory=os.path.join(dir,'MESH')   
        else:
            directory=os.path.join(os.getcwd(),'MESH')
        # create directory if not present
        if not os.path.exists(directory):                        
            os.mkdir(directory) 

        self.path=os.path.join(directory,ut.get_filename(self.subject.toPython()) + extension)

        #write files
        if o3d.io.write_triangle_mesh(self.path, self.resource):
            return True
        return False

    def get_metadata_from_resource(self) -> bool:
        """Returns the metadata from a resource. \n

        Features:
            PointCount\n
            faceCount \n
            orientedBoundingBox\n
            cartesianTransform\n
            cartesianBounds\n
            orientedBounds \n

        Returns:
            bool: True if exif data is successfully parsed
        """
        if (not self.resource or
            len(self.resource.triangles) <2):
            return False    

        try:
            if getattr(self,'pointCount',None) is None:
                self.pointCount=len(self.resource.vertices)

            if getattr(self,'faceCount',None) is None:
                self.faceCount=len(self.resource.triangles)

            if  getattr(self,'cartesianTransform',None) is None:
                center=self.resource.get_center()  
                self.cartesianTransform= np.array([[1,0,0,center[0]],
                                                    [0,1,0,center[1]],
                                                    [0,0,1,center[2]],
                                                    [0,0,0,1]])

            if getattr(self,'cartesianBounds',None) is  None:
                self.cartesianBounds=gt.get_cartesian_bounds(self.resource)
            if getattr(self,'orientedBoundingBox',None) is  None:
                self.orientedBoundingBox=self.resource.get_oriented_bounding_box()
            if getattr(self,'orientedBounds',None) is  None:
                box=self.resource.get_oriented_bounding_box()
                self.orientedBounds= np.asarray(box.get_box_points())
            return True
        except:
            raise ValueError('Metadata extraction from resource failed')
       
    



import copy
import math
import os
import shutil
import time
import unittest
from multiprocessing.sharedctypes import Value

import cv2
import geomapi.utils as ut
import geomapi.utils.geometryutils as gt
import ifcopenshell
import numpy as np
import open3d as o3d
import PIL
import pye57
import rdflib
from geomapi.nodes import *
from ifcopenshell.util.selector import Selector
from rdflib import RDF, RDFS, Graph, Literal, URIRef


class TestNode(unittest.TestCase):

################################## SETUP/TEARDOWN CLASS ######################
    @classmethod
    def setUpClass(cls):
        #execute once before all tests
        print('-----------------Setup Class----------------------')
        st = time.time()
        cls.path=os.path.join(os.getcwd(), "test","testfiles") 

        #ONTOLOGIES
        cls.exif = rdflib.Namespace('http://www.w3.org/2003/12/exif/ns#')
        cls.geo=rdflib.Namespace('http://www.opengis.net/ont/geosparql#') #coordinate system information
        cls.gom=rdflib.Namespace('https://w3id.org/gom#') # geometry representations => this is from mathias
        cls.omg=rdflib.Namespace('https://w3id.org/omg#') # geometry relations
        cls.fog=rdflib.Namespace('https://w3id.org/fog#')
        cls.v4d=rdflib.Namespace('https://w3id.org/v4d/core#')
        cls.openlabel=rdflib.Namespace('https://www.asam.net/index.php?eID=dumpFile&t=f&f=3876&token=413e8c85031ae64cc35cf42d0768627514868b2f#')
        cls.e57=rdflib.Namespace('http://libe57.org#')
        cls.xcr=rdflib.Namespace('http://www.w3.org/1999/02/22-rdf-syntax-ns#')
        cls.ifc=rdflib.Namespace('http://ifcowl.openbimstandards.org/IFC2X3_Final#')
        
        #IFC1 (IfcWall)
        cls.bimGraphPath1=os.path.join(cls.path,'bimGraph1.ttl')
        cls.bimGraph1=Graph().parse(cls.bimGraphPath1)
        cls.ifcPath1=os.path.join(cls.path,'IFC',"Academiestraat_building_1.ifc")  
        cls.ifc1=ifcopenshell.open(cls.ifcPath1)   
        cls.ifcElement1= cls.ifc1.by_guid('0KysUSO6T3_gOJKtAiUE7d')
        cls.path1=os.path.join(cls.path,'BIM',"Basic_Wall_211_WA_Ff1_Glued_brickwork_sandlime_150mm_1118860_0KysUSO6T3_gOJKtAiUE7d.ply")  
        cls.mesh1=gt.ifc_to_mesh(cls.ifcElement1)
        cls.subject1=next(s for s in cls.bimGraph1.subjects(RDF.type) if "file:///Basic_Wall_211_WA_Ff1_Glued_brickwork_sandlime_150mm_1118860_0KysUSO6T3_gOJKtAiUE7d" in s.toPython())
        
        #IFC2 (IfcSite)
        cls.bimGraphPath2=os.path.join(cls.path,'bimGraph2.ttl')
        cls.bimGraph2=Graph().parse(cls.bimGraphPath2)
        cls.ifcPath2=os.path.join(cls.path,'IFC',"Mariakerke_AWV_Conform_3D_BT_l72.ifc")  
        cls.ifc2=ifcopenshell.open(cls.ifcPath2)   
        cls.ifcElement2= cls.ifc2.by_guid('3dzlFaOIb0bx0z6uxD96Sz')
        cls.path2=os.path.join(cls.path,'BIM',"BT1_Bodembedekking_WSV10_3dzlFaOIb0bx0z6uxD96Sz.ply")  
        cls.mesh2=gt.ifc_to_mesh(cls.ifcElement2) 
        cls.subject2=next(s for s in cls.bimGraph2.subjects(RDF.type) if "file:///BT1_Bodembedekking_WSV10_3dzlFaOIb0bx0z6uxD96Sz" in s.toPython())
        
        #IFC3 (IfcSlab)
        cls.bimGraphPath3=os.path.join(cls.path,'bimGraph3.ttl')
        cls.bimGraph3=Graph().parse(cls.bimGraphPath3)
        cls.ifcPath3=os.path.join(cls.path,'IFC',"Academiestraat_parking.ifc")  
        cls.ifc3=ifcopenshell.open(cls.ifcPath3)   
        cls.ifcElement3= cls.ifc3.by_guid('3fuhig3Pv7AexdYIkSgcdp')
        cls.path3=os.path.join(cls.path,'BIM',"174_SFO_Pile_type_1_CS_800kN_TS_250kN_1341745_3fuhig3Pv7AexdYIkSgcdp.ply")  
        cls.mesh3=gt.ifc_to_mesh(cls.ifcElement3) 
        cls.subject3=next(s for s in cls.bimGraph3.subjects(RDF.type) if "file:///174_SFO_Pile_type_1_CS_800kN_TS_250kN_1341745_3fuhig3Pv7AexdYIkSgcdp" in s.toPython())
        
        #IFC4 (IfcDoor)
        cls.bimGraphPath4=os.path.join(cls.path,'bimGraph4.ttl')
        cls.bimGraph4=Graph().parse(cls.bimGraphPath4)
        cls.ifcPath4=os.path.join(cls.path,'IFC',"B1_ALG_Model.ifc")  
        cls.ifc4=ifcopenshell.open(cls.ifcPath4)   
        cls.ifcElement4= cls.ifc4.by_guid('3M6cRe8S51TfU811R8bxHi')
        cls.path4=os.path.join(cls.path,'BIM',"2RA_doors_single_omlijsting_omlijsting_-_85_x_217_1855492_3M6cRe8S51TfU811R8bxHi.ply")  
        cls.mesh4=gt.ifc_to_mesh(cls.ifcElement4) 
        cls.subject4=next(s for s in cls.bimGraph4.subjects(RDF.type) if "file:///2RA_doors_single_omlijsting_omlijsting_-_85_x_217_1855492_3M6cRe8S51TfU811R8bxHi" in s.toPython())
       
        #RESOURCES
        cls.resourcePath=os.path.join(cls.path,"resources")
        if not os.path.exists(cls.resourcePath):
            os.mkdir(cls.resourcePath)
   
        #TIME TRACKING           
        et = time.time()
        print("startup time: "+str(et - st))
        print('{:50s} {:5s} '.format('tests','time'))
        print('------------------------------------------------------')


    @classmethod
    def tearDownClass(cls):
        #execute once after all tests
        print('-----------------TearDown Class----------------------')
        if os.path.exists(cls.resourcePath):
            shutil.rmtree(cls.resourcePath)  
        if os.path.exists(os.path.join(os.getcwd(),'BIM') ):
            shutil.rmtree(os.path.join(os.getcwd(),'BIM') )
        if os.path.exists(os.path.join(os.getcwd(),'resources') ):
            shutil.rmtree(os.path.join(os.getcwd(),'resources') )

################################## SETUP/TEARDOWN ######################

    def setUp(self):
        #execute before every test
        self.startTime = time.time()   

    def tearDown(self):
        #execute after every test
        t = time.time() - self.startTime
        print('{:50s} {:5s} '.format(self._testMethodName,str(t)))
################################## TEST FUNCTIONS ######################

    def test_bimnode_creation_from_subject(self):
        #subject
        subject='myNode'
        node= BIMNode(subject=subject)
        self.assertEqual(node.subject.toPython(),'file:///'+subject)

        #http://
        subject='http://session_2022_05_20'
        node= BIMNode(subject=subject)
        self.assertEqual(node.subject.toPython(),subject)
        
        #erroneous char       
        subject='[[http://ses>sion_2022_<05_20]]'
        node= BIMNode(subject=subject)
        self.assertEqual(node.subject.toPython(),'http://__ses_sion_2022__05_20__')

    def test_bimnode_creation_from_graph(self):
        subject=next(self.bimGraph1.subjects(RDF.type))
        node= BIMNode(graph=self.bimGraph1, subject=subject)
        self.assertEqual(node.subject.toPython(),subject.toPython())
        object=self.bimGraph1.value(subject,self.v4d['faceCount'])
        self.assertEqual(node.faceCount,object.toPython())
        
    def test_bimnode_creation_from_graph_path(self):
        subject=next(self.bimGraph2.subjects(RDF.type))
        node= BIMNode(graphPath=self.bimGraphPath2, subject=subject)
        self.assertEqual(node.subject.toPython(),subject.toPython())
        object=self.bimGraph2.value(subject,self.v4d['faceCount'])
        self.assertEqual(node.faceCount,object.toPython())
        
    def test_bimnode_creation_from_path(self):
        #path1 without getResource
        node= BIMNode(path=self.path1)
        self.assertEqual(node.name,ut.get_filename(self.path1))
        self.assertIsNone(node.resource)

        #path2 with getResource
        node= BIMNode(path=self.path2,getResource=True)        
        self.assertEqual(node.name,ut.get_filename(self.path2))
        self.assertEqual(node.faceCount,len(self.mesh2.triangles))

        #path3 
        node= BIMNode(path=self.path3,getResource=True)
        self.assertEqual(node.name,ut.get_filename(self.path3))
        self.assertEqual(node.faceCount,len(self.mesh3.triangles))

    def test_bimnode_creation_from_ifcpath(self):
        #path1 without extra info -> take first IfcElement
        node= BIMNode(ifcPath=self.ifcPath3)
        self.assertIsNotNone(node.name)

        #path2 with getResource without extra info -> take first IfcElement
        node= BIMNode(ifcPath=self.ifcPath4,getResource=True)        
        self.assertIsNotNone(node.name)
        self.assertIsNotNone(node.resource)

 
    def test_bimnode_creation_from_ifcElement(self):        
        #ifcElement
        node= BIMNode(resource=self.ifcElement1)
        self.assertEqual(node.subject.toPython(),self.subject1.toPython())

        #ifcElement + getResource
        node= BIMNode(resource=self.ifcElement2,ifcPath=self.ifcPath2)
        self.assertEqual(node.subject,self.subject2)
        self.assertIsNotNone(node.resource)
        self.assertIsNotNone(node.ifcPath)

    def test_bimnode_creation_from_mesh(self):
        #mesh1
        node= BIMNode(resource=self.mesh1)
        self.assertEqual(node.faceCount,len(self.mesh1.triangles))
        #mesh2
        node= BIMNode(resource=self.mesh2)
        self.assertEqual(node.faceCount,len(self.mesh2.triangles))
        #mesh3
        node= BIMNode(resource=self.mesh3)
        self.assertEqual(node.faceCount,len(self.mesh3.triangles))

    def test_bimnode_creation_from_trimesh(self):
        myTrimesh=gt.mesh_to_trimesh(self.mesh2)
        node= BIMNode(resource=myTrimesh)
        self.assertEqual(node.faceCount,len(self.mesh2.triangles))

    def test_creation_from_subject_and_graph_and_graphPath(self):        
        subject=next(self.bimGraph3.subjects(RDF.type))
        node= BIMNode(subject=subject,graph=self.bimGraph3,graphPath=self.bimGraphPath3)
        self.assertEqual(node.subject.toPython(),subject.toPython())
        node.to_graph()
        self.assertTrue((subject, self.v4d['faceCount'], Literal(node.faceCount)) in self.bimGraph3)
   
    def test_creation_from_subject_and_path(self):        
        node= BIMNode(subject=self.subject2,path=self.path2,getResource=True)
        self.assertEqual(node.subject.toPython(),self.subject2.toPython())
        node.to_graph()        
        self.assertAlmostEqual(np.asarray(node.orientedBounds)[0,0],np.asarray(self.mesh2.get_oriented_bounding_box().get_box_points())[0,0],delta=0.01)
        
    def test_creation_from_subject_and_path_and_graph(self):      
        node= BIMNode(subject=self.subject3,path=self.path3,graph=self.bimGraph3,getResource=True)
        self.assertEqual(node.subject,self.subject3)
        node.to_graph()
        initialGraph=ut.get_subject_graph(self.bimGraph3,subject=self.subject3)
        self.assertAlmostEqual(node.cartesianBounds[0],self.mesh3.get_min_bound()[0],delta=0.01)

    def test_creation_from_mesh_and_path (self):      
        node= BIMNode(resource=self.mesh4,path=self.path4)
        self.assertEqual(node.subject,self.subject4)

    def test_creation_from_subject_and_mesh_and_path_and_graph(self):      
        node= BIMNode(subject=self.subject3,resource=self.mesh3,path=self.path3, graph=self.bimGraph3)
        self.assertEqual(node.subject,self.subject3)
        node.to_graph()
        object=node.graph.value(node.subject,self.v4d['path'])
        self.assertEqual(object.toPython(),os.path.join("BIM",ut.get_filename(self.path3)+'.ply') )

    def test_creation_from_subject_mesh_path_ifcPath_graph(self):      
        node= BIMNode(subject=self.subject4,resource=self.mesh4,path=self.path4, ifcPath=self.ifcPath4, graph=self.bimGraph4)
        self.assertEqual(node.subject,self.subject4 )
        node.to_graph()
        object=node.graph.value(node.subject,self.v4d['path'])
        self.assertEqual(object.toPython(),os.path.join("BIM",ut.get_filename(self.path4)+'.ply') )

    def test_creation_from_subject_mesh_path_ifcPath_globalId_graph(self):      
        node= BIMNode(subject=self.subject1,resource=self.mesh1,path=self.path1, ifcPath=self.ifcPath1, globalId=self.ifcElement1.GlobalId, graph=self.bimGraph1)
        self.assertEqual(node.subject,self.subject1)
        node.to_graph()
        object=node.graph.value(node.subject,self.v4d['path'])
        self.assertEqual(object.toPython(),os.path.join("BIM",ut.get_filename(self.path1)+'.ply') )

    def test_node_creation_with_get_resource(self):
        #mesh
        node= BIMNode(resource=self.mesh1)
        self.assertIsNotNone(node.resource)

        #path without getResource
        node= BIMNode(path=self.path2)
        self.assertIsNone(node.resource)

        #path with getResource
        node= BIMNode(path=self.path3,getResource=True)
        self.assertIsNotNone(node.resource)

        #graph with get resource
        node= BIMNode(subject=self.subject2,graph=self.bimGraph2,getResource=True)
        self.assertIsNone(node.resource)
        
        #graphPath with get resource
        node= BIMNode(subject=self.subject3,graphPath=self.bimGraphPath3,getResource=True)
        self.assertIsNotNone(node.resource)

    def test_clear_resource(self):
        #mesh
        node= BIMNode(resource=self.mesh1)
        self.assertIsNotNone(node.resource)
        del node.resource
        self.assertIsNone(node.resource)

    # def test_save_resource(self):
    #     #no mesh -> False
    #     node= BIMNode()
    #     self.assertFalse(node.export_resource())

    #     #directory
    #     node= BIMNode(mesh=self.mesh2)
    #     self.assertTrue(node.export_resource(os.path.join(self.path,'resources')))

    #     #graphPath        
    #     node= BIMNode(mesh=self.mesh2,graphPath=self.bimGraphPath2)
    #     self.assertTrue(node.export_resource())

    #     #no path or graphPath
    #     node= BIMNode(mesh=self.mesh4)        
    #     self.assertTrue(node.export_resource())

    #     #invalid extension -> error
    #     node= BIMNode(mesh=self.mesh3)
    #     self.assertRaises(ValueError,node.export_resource,os.path.join(self.path,'resources'),'.kjhgfdfg')

    #     #.ply -> ok
    #     node= BIMNode(mesh=self.mesh2)
    #     self.assertTrue(node.export_resource(os.path.join(self.path,'resources'),'.ply'))
    #     self.assertEqual(node.path,os.path.join(self.path,'resources',node.name+'.ply'))

    #     #.obj -> ok
    #     node= BIMNode(mesh=self.mesh3)
    #     self.assertTrue(node.export_resource(os.path.join(self.path,'resources'),'.obj'))
    #     self.assertEqual(node.path,os.path.join(self.path,'resources',node.name+'.obj'))
        
    #     #path -> new name
    #     node= BIMNode(subject=URIRef('myMesh'),path=self.path2,getResource=True)
    #     self.assertTrue(node.export_resource())
    #     files=ut.get_list_of_files(ut.get_folder(node.path))
    #     self.assertTrue( node.path in files)
        
    #     #graphPath with directory
    #     node=BIMNode(subject=self.subject2,graphPath=self.bimGraphPath2, mesh=self.mesh3)
    #     self.assertTrue(node.export_resource(os.path.join(self.path,'resources')))
    #     files=ut.get_list_of_files(ut.get_folder(node.path))
    #     self.assertTrue(node.path in files)

    #     #graph with new subject
    #     node=BIMNode(subject=self.subject4,grap=self.bimGraph4, mesh=self.mesh4)
    #     node.name='myMesh'
    #     self.assertTrue(node.export_resource())
    #     files=ut.get_list_of_files(ut.get_folder(node.path))
    #     self.assertTrue(node.path in files)

    def test_get_resource(self):
        #mesh
        node=BIMNode(resource=self.mesh3)  
        self.assertIsNotNone(node.get_resource())

        #no mesh
        del node.resource
        self.assertIsNone(node.get_resource())

        #graphPath with getResource
        node=BIMNode(graphPath=self.bimGraphPath2,subject=self.subject2,getResource=True)
        self.assertIsNotNone(node.get_resource())

    def test_get_metadata_from_resource(self):
        #mesh
        node=BIMNode(resource=self.mesh3)  
        self.assertIsNotNone(node.orientedBounds)
        self.assertIsNotNone(node.cartesianBounds)
        self.assertIsNotNone(node.cartesianTransform)
        self.assertIsNotNone(node.faceCount)
        self.assertIsNotNone(node.pointCount)

        #ifcElement
        node=BIMNode(resource=self.ifcElement4)  
        self.assertIsNotNone(node.orientedBounds)
        self.assertIsNotNone(node.cartesianBounds)
        self.assertIsNotNone(node.cartesianTransform)
        self.assertIsNotNone(node.faceCount)
        self.assertIsNotNone(node.pointCount)

        #ifcPath and globalId
        node=BIMNode(ifcPath=self.ifcPath1,globalId=self.ifcElement1.GlobalId, getResource=True)  
        self.assertIsNotNone(node.orientedBounds)
        self.assertIsNotNone(node.cartesianBounds)
        self.assertIsNotNone(node.cartesianTransform)
        self.assertIsNotNone(node.faceCount)
        self.assertIsNotNone(node.pointCount)

        #graphPath
        node=BIMNode(graphPath=self.bimGraphPath4,subject=self.subject4,getResource=True)
        self.assertIsNotNone(node.orientedBounds)
        self.assertIsNotNone(node.cartesianBounds)
        self.assertIsNotNone(node.cartesianTransform)
        self.assertIsNotNone(node.faceCount)
        self.assertIsNotNone(node.pointCount)

    def test_set_path(self):
        #valid path
        node=BIMNode()
        node.path= self.path1
        self.assertEqual(node.path,self.path1)

        #preexisting
        node=BIMNode(path=self.path4)
        self.assertEqual(node.path,self.path4)

        #graphPath & name
        node=BIMNode(subject=self.subject4,graphPath=self.bimGraphPath4)
        node.get_path()
        self.assertEqual(node.path,self.path4)

    def test_set_ifc_path(self):
        #valid path
        node=BIMNode(ifcPath=self.ifcPath2)
        self.assertEqual(node.ifcPath,self.ifcPath2)

        #invalid
        self.assertRaises(ValueError,BIMNode,ifcPath='qsffqsdf.dwg')

    def get_metadata_from_ifc_path(self):
        #ifc1
        a=0
        selector = Selector()
        for ifcElement in selector.parse(self.ifc1, '.ifcObject'):
            a+=1
            node=BIMNode(ifcPath=self.ifcPath1,globalId=ifcElement.GlobalId)            
            self.assertEqual(node.className,ifcElement.is_a())
            if a==100:
                break
        
        #check error no global id in ifc
        node= BIMNode()
        self.assertRaises(ValueError,BIMNode,ifcPath=self.ifcPath4,globalId='kjhgfdfg')

    def test_get_metadata_from_ifcElement(self):
        #ifc3
        a=0
        selector = Selector()
        for ifcElement in selector.parse(self.ifc3, '.ifcObject'):
            a+=1
            node=BIMNode(resource=ifcElement)
            self.assertEqual(node.className,ifcElement.is_a())
            if a==100:
                break

        #ifc4
        a=0
        selector = Selector()
        for ifcElement in selector.parse(self.ifc4, '.ifcObject'):
            a+=1
            node=BIMNode(resource=ifcElement)
            self.assertEqual(node.className,ifcElement.is_a())
            if a==100:
                break

if __name__ == '__main__':
    unittest.main()



"""
linkeddatatools - a Python library for RDF graph structuring and exchange.
"""
#IMPORT PACKAGES
from dis import dis
from lib2to3.pytree import Node
import numpy as np 
import cv2 
import open3d as o3d 
import os 
import re
import pye57 #conda install xerces-c  =>  pip install pye57

import xml.etree.ElementTree as ET 
from typing import List,Tuple

# import APIs
import rdflib
from rdflib import Graph, plugin
from rdflib.serializer import Serializer #pip install rdflib-jsonld https://pypi.org/project/rdflib-jsonld/
from rdflib import Graph
from rdflib import URIRef, BNode, Literal
from rdflib.namespace import CSVW, DC, DCAT, DCTERMS, DOAP, FOAF, ODRL2, ORG, OWL, \
                           PROF, PROV, RDF, RDFS, SDO, SH, SKOS, SOSA, SSN, TIME, \
                           VOID, XMLNS, XSD
import ifcopenshell
import ifcopenshell.geom as geom
import ifcopenshell.util
from ifcopenshell.util.selector import Selector
import multiprocessing
import concurrent.futures

#IMPORT MODULES 
from geomapi.nodes import *
from geomapi.nodes.sessionnode import create_node 
import geomapi.utils as ut
import geomapi.utils.geometryutils as gt

from warnings import warn

#### NODE CREATION ####

def e57xml_to_nodes(e57XmlPath :str, **kwargs) -> List[PointCloudNode]:
    """Parse XML file that is created with E57lib e57xmldump.exe.

    Args:
        path (string):  e57 xml file path e.g. "D:\\Data\\2018-06 Werfopvolging Academiestraat Gent\\week 22\\PCD\\week 22 lidar_CC.xml"
            
    Returns:
        A list of pointcloudnodes with the xml metadata 
    """
    if os.path.exists(e57XmlPath) and e57XmlPath.endswith('.xml'):    
        #E57 XML file structure
        #e57Root
        #   >data3D
        #       >vectorChild
        #           >pose
        #               >rotation
        #               >translation
        #           >cartesianBounds
        #           >guid
        #           >name
        #           >points recordCount
        #   >images2D
        mytree = ET.parse(e57XmlPath)
        root = mytree.getroot()  
        nodelist=[]   
        e57Path=e57XmlPath.replace('.xml','.e57')       

        for idx,child in enumerate(root.iter('{http://www.astm.org/COMMIT/E57/2010-e57-v1.0}vectorChild')):
            nodelist.append(PointCloudNode(e57XmlPath=e57XmlPath,e57Index=idx,path=e57Path,**kwargs))
        return nodelist
    else:
        raise ValueError('No valid e57XmlPath.')

def img_xml_to_nodes(xmlPath :str, **kwargs) -> List[ImageNode]:
    """Parse XML file that is created with https://www.agisoft.com/.

    Args:
        path (string):  e57 xml file path e.g. "D:\\Data\\2018-06 Werfopvolging Academiestraat Gent\\week 22\\PCD\\week 22 lidar_CC.xml"
            
    Returns:
        A list of pointcloudnodes with the xml metadata 
    """
    if os.path.exists(xmlPath) and xmlPath.endswith('.xml'):    
        mytree = ET.parse(xmlPath)
        root = mytree.getroot()  
        nodelist=[]   
        for child in root.iter('camera'):
            nodelist.append(ImageNode(xmlPath=xmlPath,name=child.get('label'),**kwargs))
        return nodelist
    else:
        raise ValueError('No valid xmlPath.')

def e57path_to_nodes(e57Path:str,percentage:float=1.0) ->List[PointCloudNode]:
    """Load an e57 file and convert all data to a list of PointCloudNodes.\n

    **NOTE**: lowering the percentage barely affects assignment performance (numpy array assignments are extremely efficient). \n 
    Only do this to lower computational complexity or free up memory.

    Args:
        1. e57path(str): absolute path to .e57 file\n
        2. percentage(float,optional): percentage of points to load. Defaults to 1.0 (100%)\n

    Returns:
        o3d.geometry.PointCloud
    """    
    e57 = pye57.E57(e57Path)
    nodes=[]
    for s in range(e57.scan_count):
        resource=gt.e57_to_pcd(e57,e57Index=s,percentage=percentage)
        node=PointCloudNode(resource=resource,
                            path=e57Path,
                            e57Index=s,
                            percentage=percentage)
        node.pointCount=len(resource.points)
        nodes.append(node)
    return nodes
    
def e57path_to_nodes_mutiprocessing(e57Path:str,percentage:float=1.0) ->List[PointCloudNode]:
    """Load an e57 file and convert all data to a list of PointCloudNodes.\n

    **NOTE**: Complex types cannot be pickled (serialized) by Windows. Therefore, a two step parsing is used where e57 data is first loaded as np.arrays with multi-processing.
    Next, the arrays are passed to o3d.geometry.PointClouds outside of the loop.\n  

    **NOTE**: starting parallel processing takes a bit of time. This method will start to outperform single-core import from 3+ pointclouds.\n

    **NOTE**: lowering the percentage barely affects assignment performance (numpy array assignments are extremely efficient). \n 
    Only do this to lower computational complexity or free up memory.

    Args:
        1. e57path(str): absolute path to .e57 file\n
        2. percentage(float,optional): percentage of points to load. Defaults to 1.0 (100%)\n

    Returns:
        o3d.geometry.PointCloud
    """    
    e57 = pye57.E57(e57Path)
    nodes=[]
    with concurrent.futures.ProcessPoolExecutor() as executor:
        # first load all e57 data and output it as np.arrays
        results=[executor.submit(gt.e57_to_arrays,e57Path=e57Path,e57Index=s,percentage=percentage) for s in range(e57.scan_count)]
        # next, the arrays are assigned to point clouds outside the loop.
        for s,r in enumerate(concurrent.futures.as_completed(results)):
            resource=gt.arrays_to_pcd(r.result())
            node=PointCloudNode(resource=resource,
                                path=e57Path,
                                e57Index=s,
                                percentage=percentage)
            node.pointCount=len(resource.points)
            nodes.append(node)
    return nodes

def e57header_to_nodes(e57Path:str, **kwargs) -> List[PointCloudNode]:
    """
    Parse e57 file header that is created with E57lib e57xmldump.exe.

    Args:
        path (string):  e57 xml file path e.g. "D:\\Data\\2018-06 Werfopvolging Academiestraat Gent\\week 22\\PCD\\week 22 lidar_CC.xml"
            
    Returns:
        A list of pointcloudnodes with the xml metadata 
    """
    if os.path.exists(e57Path) and e57Path.endswith('.e57'):    

        nodelist=[]   
        e57 = pye57.E57(e57Path)   
        for idx in range(e57.scan_count):
            nodelist.append(PointCloudNode(path=e57Path,e57Index=idx, **kwargs))
        return nodelist
    else:
        raise ValueError('No valid e57Path.')

def ifc_to_nodes(ifcPath:str, classes:str='.IfcBuildingElement',getResource : bool=True,**kwargs)-> List[BIMNode]:
    """
    Parse ifc file to a list of BIMNodes, one for each ifcElement.\n

    **NOTE**: classes are not case sensitive. It is advised to solely focus on IfcBuildingElement classes or inherited classes as these typically have geometry representations that can be used by GEOMAPI.

    **NOTE**: If you intend to parse 1000+ elements, use the multithreading of the entire file instead and filter the BIMNodes afterwards as it will be faster. 

    **WARNING**: IfcOpenShell strugles with some ifc serializations. In our experience, IFC4 serializations is more robust.

    .. image:: ../docs/pics/ifc_inheritance.PNG

    Args:
        1. ifcPath (string):  absolute ifc file path e.g. "D:\\myifc.ifc"\n
        2. classes (string, optional): ifcClasses seperated by | e.g. '.IfcBeam | .IfcColumn '#'.IfcWall | .IfcSlab | .IfcBeam | .IfcColumn | .IfcStair | .IfcWindow | .IfcDoor'. Defaults to '.IfcBuildingElement'.   
    
    Raises:
        ValueError: 'No valid ifcPath.'

    Returns:
        List[BIMNode]
    """   
    if os.path.exists(ifcPath) and ifcPath.endswith('.ifc'):    
        nodelist=[]   
        ifc = ifcopenshell.open(ifcPath)   
        selector = Selector()
        for ifcElement in selector.parse(ifc, classes):
            # print(ifcElement.Name) # THIS IS TEMP
            node=BIMNode(resource=ifcElement,getResource=getResource, **kwargs)          
            node.ifcPath=ifcPath
            nodelist.append(node)
        return nodelist
    else:
        raise ValueError('No valid ifcPath.')


def ifc_to_nodes_by_type(ifcPath:str, types:list=['IfcBuildingElement'],getResource : bool=True,**kwargs)-> List[BIMNode]:
    """
    Parse ifc file to a list of BIMNodes, one for each ifcElement.\n

    **NOTE**: classes are not case sensitive. It is advised to solely focus on IfcBuildingElement classes or inherited classes as these typically have geometry representations that can be used by GEOMAPI.

    **WARNING**: IfcOpenShell strugles with some ifc serializations. In our experience, IFC4 serializations is more robust.

    .. image:: ../docs/pics/ifc_inheritance.PNG

    Args:
        1. ifcPath (string):  absolute ifc file path e.g. "D:\\myifc.ifc"\n
        2. types (list of strings, optional): ifcClasses you want to parse e.g. ['IfcWall','IfcSlab','IfcBeam','IfcColumn','IfcStair','IfcWindow','IfcDoor']. Defaults to ['IfcBuildingElement']. \n  
    
    Raises:
        ValueError: 'No valid ifcPath.'

    Returns:
        List[BIMNode]
    """   
    #validate types

    if os.path.exists(ifcPath) and ifcPath.endswith('.ifc'):    
        try:
            ifc_file = ifcopenshell.open(ifcPath)
        except:
            print(ifcopenshell.get_log())
        else:
            nodelist=[]   
            for type in types:
                ifcElements = ifc_file.by_type(type)
                for ifcElement in ifcElements:
                    node=BIMNode(resource=ifcElement,getResource=getResource, **kwargs)          
                    node.ifcPath=ifcPath
                    nodelist.append(node)
            return nodelist
    else:
        raise ValueError('No valid ifcPath.')

def ifc_to_nodes_multiprocessing(ifcPath:str, **kwargs)-> List[BIMNode]:
    """Returns the contents of geometry elements in an ifc file as BIMNodes.\n
    This method is 3x faster than other parsing methods due to its multi-threading.\n
    However, only the entire ifc can be parsed.\n

    **WARNING**: IfcOpenShell strugles with some ifc serializations. In our experience, IFC4 serializations is more robust.


    Args:
        ifcPath (str): path (string):  absolute ifc file path e.g. "D:\\myifc.ifc"\n

    Raises:
        ValueError: 'No valid ifcPath.'

    Returns:
        List[BIMNode]
    """
    if os.path.exists(ifcPath) and ifcPath.endswith('.ifc'):  
        try:
            ifc_file = ifcopenshell.open(ifcPath)
        except:
            print(ifcopenshell.get_log())
        else: 
            nodelist=[]   
            timestamp=ut.get_timestamp(ifcPath)
            settings = ifcopenshell.geom.settings()
            settings.set(settings.USE_WORLD_COORDS, True) 
            iterator = ifcopenshell.geom.iterator(settings, ifc_file, multiprocessing.cpu_count())
            if iterator.initialize():
                while True:
                    shape = iterator.get()
                    ifcElement = ifc_file.by_guid(shape.guid) 
                    faces = shape.geometry.faces # Indices of vertices per triangle face e.g. [f1v1, f1v2, f1v3, f2v1, f2v2, f2v3, ...]
                    verts = shape.geometry.verts # X Y Z of vertices in flattened list e.g. [v1x, v1y, v1z, v2x, v2y, v2z, ...]
                    # materials = shape.geometry.materials # Material names and colour style information that are relevant to this shape
                    # material_ids = shape.geometry.material_ids # Indices of material applied per triangle face e.g. [f1m, f2m, ...]

                    # Since the lists are flattened, you may prefer to group them per face like so depending on your geometry kernel
                    grouped_verts = [[verts[i], verts[i + 1], verts[i + 2]] for i in range(0, len(verts), 3)]
                    grouped_faces = [[faces[i], faces[i + 1], faces[i + 2]] for i in range(0, len(faces), 3)]

                    #Convert grouped vertices/faces to Open3D objects 
                    o3dVertices = o3d.utility.Vector3dVector(np.asarray(grouped_verts))
                    o3dTriangles = o3d.utility.Vector3iVector(np.asarray(grouped_faces))

                    # Create the Open3D mesh object
                    mesh=o3d.geometry.TriangleMesh(o3dVertices,o3dTriangles)

                    #if mesh, create node
                    if len(mesh.triangles)>1:
                        node=BIMNode(**kwargs)
                        node.name=ifcElement.Name
                        node.className=ifcElement.is_a()
                        node.globalId=ifcElement.GlobalId
                        if node.name and node.globalId:
                            node.subject= node.name +'_'+node.globalId 
                        node.resource=mesh
                        node.get_metadata_from_resource()
                        node.timestamp=timestamp
                        node.ifcPath=ifcPath
                        nodelist.append(node)
                        
                    if not iterator.next():
                        break
            return nodelist
    else:
        raise ValueError('No valid ifcPath.') 


##### NODE SELECTION #####

def select_k_nearest_nodes(node:Node,nodelist:List[Node],k:int=10) -> Tuple[List [Node], o3d.utility.DoubleVector]:
    """ Select k nearest nodes based on Euclidean distance between centroids.\n

    .. image:: ../docs/pics/selection_k_nearest.PNG

    Args:
        0. node (Node): node to search from\n
        1. nodelist (List[Node])\n
        2. k (int, optional): number of neighbors. Defaults to 10.\n

    Returns:
        List of Nodes
    """
    if k <=0:
        raise ValueError('k must be positive and non-negative.')

    #get node center
    if node.get_cartesian_transform() is not None:
        point=gt.get_translation(node.cartesianTransform)
        #create pcd from nodelist centers
        pcd = o3d.geometry.PointCloud()
        array=np.empty(shape=(len(nodelist),3))
        for idx,node in enumerate(nodelist):
            if node.get_cartesian_transform() is not None:
                array[idx]=gt.get_translation(node.cartesianTransform)
            else:
                array[idx]=[-10000.0,-10000.0,-10000.0]
        pcd.points = o3d.utility.Vector3dVector(array)

        #Create KDTree from pcd
        pcdTree = o3d.geometry.KDTreeFlann(pcd)

        #Find 200 nearest neighbors
        _, idxList, distances = pcdTree.search_knn_vector_3d(point, k)
        selectedNodeList=[node for idx,node in enumerate(nodelist) if idx in idxList]

        if any(selectedNodeList):        
            return selectedNodeList, distances
    else:
        return None,None

def select_nodes_with_centers_in_radius(node:Node,nodelist:List[Node],r:float=0.5) -> List [Node]:
    """Select nodes within radius of the node centroid based on Euclidean distance between node centroids.\n

    .. image:: ../docs/pics/selection_radius_nearest.PNG
    
    Args:
        0. node (Node): node to search from\n
        1. nodelist (List[Node])\n
        2. r (float, optional): radius to search. Defaults to 0.5m.\n

    Returns:
        List of Nodes
    """
    
    if r <=0:
        raise ValueError('r must be positive and non-negative.')
    #get node center
    if node.get_cartesian_transform() is not None:
        point=gt.get_translation(node.cartesianTransform)
        #create pcd from nodelist centers
        pcd = o3d.geometry.PointCloud()
        array=np.empty(shape=(len(nodelist),3))
        for idx,node in enumerate(nodelist):
            if node.get_cartesian_transform() is not None:
                array[idx]=gt.get_translation(node.cartesianTransform)
            else:
                array[idx]=[-10000.0,-10000.0,-10000.0]
        pcd.points = o3d.utility.Vector3dVector(array)

        #Create KDTree from pcd
        pcdTree = o3d.geometry.KDTreeFlann(pcd)

        #Find 200 nearest neighbors
        [_, idxList, distances] = pcdTree.search_radius_vector_3d(point, r)
        selectedNodeList=[node for idx,node in enumerate(nodelist) if idx in idxList ]
        selectedNodeList=[node for i,node in enumerate(selectedNodeList) if distances[i]<=r ]
        
        if any(selectedNodeList):        
            return selectedNodeList,distances
    else:
        return None,None

def select_nodes_with_centers_in_bounding_box(node:Node,nodelist:List[Node],u:float=0.5,v:float=0.5,w:float=0.5) -> List [Node]: 
    """Select the nodes of which the center lies within the oriented Bounding Box of the source node given an offset.\n

    .. image:: ../docs/pics/selection_box_inliers.PNG
    
    Args:
        0. node (Node): source Node \n
        1. nodelist (List[Node]): target nodelist\n
        2. u (float, optional): Offset in X. Defaults to 0.5m.\n
        3. v (float, optional): Offset in Y. Defaults to 0.5m.\n
        4. w (float, optional): Offset in Z. Defaults to 0.5m.\n

    Returns:
        List [Node]
    """
    #get box source node
    if node.get_oriented_bounding_box() is not None:
        box=node.orientedBoundingBox
        box=gt.expand_box(box,u=u,v=v,w=w)

        # get centers
        centers=np.empty((len(nodelist),3),dtype=float)
        for idx,node in enumerate(nodelist):
            if node.get_cartesian_transform() is not None:
                centers[idx]=gt.get_translation(node.cartesianTransform)

        #points are the centers of all the nodes
        pcd = o3d.geometry.PointCloud()
        points = o3d.utility.Vector3dVector(centers)
        pcd.points=points

        # Find the nodes that lie within the index box 
        idxList=box.get_point_indices_within_bounding_box(points)
        selectedNodeList=[node for idx,node in enumerate(nodelist) if idx in idxList]
        if any(selectedNodeList):        
            return selectedNodeList
    else:
        return None

def select_nodes_with_bounding_points_in_bounding_box(node:Node,nodelist:List[Node],u:float=0.5,v:float=0.5,w:float=0.5) -> List [Node]: 
    """Select the nodes of which atleast one of the bounding points lies within the oriented Bounding Box of the source node given an offset.\n

    .. image:: ../docs/pics/selection_BB_intersection.PNG
    
    Args:
        0. node (Node): source Node \n
        1. nodelist (List[Node]): target nodelist\n
        2. u (float, optional): Offset in X. Defaults to 0.5m.\n
        3. v (float, optional): Offset in Y. Defaults to 0.5m.\n
        4. w (float, optional): Offset in Z. Defaults to 0.5m.\n

    Returns:
        List [Node]
    """
    #get box source node
    if node.get_oriented_bounding_box() is not None:
        box=node.orientedBoundingBox
        box=gt.expand_box(box,u=u,v=v,w=w)

        # get boxes nodelist
        boxes=np.empty((len(nodelist),1),dtype=o3d.geometry.OrientedBoundingBox)
        for idx,node in enumerate(nodelist):
            boxes[idx]=node.get_oriented_bounding_box()

        # Find the nodes of which the bounding points lie in the source node box
        idxList=gt.get_box_inliers(box,boxes)
        selectedNodeList=[node for idx,node in enumerate(nodelist) if idx in idxList]
        if any(selectedNodeList):        
            return selectedNodeList
    else:
        return None
    
def select_nodes_with_intersecting_bounding_box(node:Node,nodelist:List[Node],u:float=0.5,v:float=0.5,w:float=0.5) -> List [Node]: 
    """Select the nodes of which the bounding boxes intersect.\n

    .. image:: ../docs/pics/selection_BB_intersection2.PNG

    Args:
        0. node (Node): source Node \n
        1. nodelist (List[Node]): target nodelist\n
        2. u (float, optional): Offset in X. Defaults to 0.5m.\n
        3. v (float, optional): Offset in Y. Defaults to 0.5m.\n
        4. w (float, optional): Offset in Z. Defaults to 0.5m.\n

    Returns:
        List [Node]
    """
    #get box source node
    if node.get_oriented_bounding_box() is not None:
        box=node.orientedBoundingBox
        box=gt.expand_box(box,u=u,v=v,w=w)

        # get boxes nodelist
        boxes=np.empty((len(nodelist),1),dtype=o3d.geometry.OrientedBoundingBox)
        for idx,node in enumerate(nodelist):
            boxes[idx]=node.get_oriented_bounding_box()
        
        # Find the nodes of which the bounding box itersects with the source node box
        idxList=gt.get_box_intersections(box,boxes)
        selectedNodeList=[node for idx,node in enumerate(nodelist) if idx in idxList]
        if any(selectedNodeList):        
            return selectedNodeList
    else:
        return None

def select_nodes_with_intersecting_resources(node:Node,nodelist:List[Node]) -> List [Node]: 
    """Select the nodes of which the o3d.geometry.TriangleMeshes intersect.\n
    This method relies on trimesh and fcl libraries for collision detection.\n
    For PointCloudNodes, the convex hull is used.\n
    For ImageNodes, a virtual mesh cone is used with respect to the field of view.\n

    .. image:: ../docs/pics/collision_5.PNG

    Args:
        0. node (Node): source Node \n
        1. nodelist (List[Node]): target nodelist\n

    Returns:
        List [Node] 
    """
    #get geometry source node
    if node.get_resource() is not None: 
        mesh=get_mesh_representation(node)
        # get geometries nodelist        
        # meshes=np.empty((len(nodelist),1),dtype=o3d.geometry.TriangleMesh)
        
        meshes=[None]*len(nodelist)
        for idx,testnode in enumerate(nodelist):
            if testnode.get_resource() is not None: 
                    meshes[idx]=get_mesh_representation(testnode)

        # Find the nodes of which the geometry itersects with the source node box
        idxList=gt.get_mesh_inliers(reference=mesh,sources=meshes)

        # idxList=gt.get_mesh_collisions_trimesh(mesh,meshes)
        selectedNodeList=[node for idx,node in enumerate(nodelist) if idx in idxList]
        if any(selectedNodeList):        
            return selectedNodeList
    return None

#### GRAPH CREATION #####

def get_mesh_representation(node: Node)->o3d.geometry.TriangleMesh:
    """Returns the mesh representation of a node resource\n
    Returns the convex hull if it is a PointCloudNode.\n
    For ImageNodes, a virtual mesh cone is used with respect to the field of view.

    Args:
        Node

    Returns:
        o3d.geometry.TriangleMesh 
    """
    nodeType=str(type(node))
    resource= node.get_resource()
   
    if 'PointCloudNode' in str(type(node)):
        hull, _ =resource.compute_convex_hull()
        return hull
    elif 'ImageNode' in nodeType:
        return node.get_mesh_geometry()
    elif 'OrthoNode' in nodeType:
        print('not implemented')
        return None
    else:
        return resource

def nodes_to_graph(nodelist : List[Node], graphPath:str =None, overwrite: bool =False,save: bool =False) -> Graph:
    """Convert list of nodes to an RDF graph.\n

    Args:
        0. nodelist (List[Node])\n
        1. graphPath (str, optional): path that serves as the basepath for all path information in the graph. This is also the storage location of the graph.\n
        2. overwrite (bool, optional): Overwrite the existing graph triples. Defaults to False.\n
        3. save (bool, optional): Save the Graph to file. Defaults to False.\n

    Returns:
        Graph 
    """
    g=Graph()
    g=ut.bind_ontologies(g)
    for node in nodelist:
            node.to_graph(graphPath,overwrite=overwrite)
            g+= node.graph
    if(graphPath and save):
        g.serialize(graphPath)     
    return g  

#### OBSOLETE #####

def graph_path_to_nodes(graphPath : str,**kwargs) -> List[Node]:
    """Convert a graphPath to a set of Nodes.

    Args:
        0. graphPath (str):  absolute path to .ttl RDF Graph\n
        1. kwargs (Any) \n

    Returns:
        A list of pointcloudnodes, imagenodes, meshnodes, bimnodes, orthonodes with metadata 
    """    
    if os.path.exists(graphPath) and graphPath.endswith('.ttl'):
        nodelist=[]
        graph=Graph().parse(graphPath)
        for subject in graph.subjects(RDF.type):
            myGraph=ut.get_subject_graph(graph,subject)
            nodelist.append(create_node(graph=myGraph,graphPath=graphPath,subject=subject,**kwargs) )
        return nodelist
    else:
        raise ValueError('No valid graphPath (only .ttl).')

def graph_to_nodes(graph : Graph,**kwargs) -> List[Node]:
    """Convert a graph to a set of Nodes.

    Args:
        0. graph (RDFlib.Graph):  Graph to parse\n
        1. kwargs (Any) \n

    Returns:
        A list of pointcloudnodes, imagenodes, meshnodes, bimnodes, orthonodes with metadata 
    """    
    nodelist=[]
    for subject in graph.subjects(RDF.type):
        node=create_node(graph=graph,subject=subject,**kwargs) 
        nodelist.append(node)
    return nodelist

# def subject_to_node_type(graph: Graph , subject:URIRef, **kwargs)-> Node:
#     # warn("This function is depricated use a SessionNode instead")

#     nodeType = ut.literal_to_string(graph.value(subject=subject,predicate=RDF.type))
#     g = Graph()
#     g += graph.triples((subject, None, None))
#     if 'BIMNode' in nodeType:
#         node=BIMNode(graph=g,**kwargs)
#     elif 'MeshNode' in nodeType:
#         node=MeshNode(graph=g,**kwargs)
#     elif 'PointCloudNode' in nodeType:
#         node=PointCloudNode(graph=g,**kwargs)
#     elif 'ImageNode' in nodeType:
#         node=ImageNode(graph=g,**kwargs)
#     elif 'SessionNode' in nodeType:
#         node=SessionNode(graph=g,**kwargs)  
#     else:
#         node=Node(graph=g,**kwargs) 
#     return node
def create_node(graph: Graph = None, graphPath: str =None, subject: URIRef = None, resource = None, **kwargs)-> Node:
    """_summary_

    Args:
        graph (Graph, optional): _description_. Defaults to None.
        graphPath (str, optional): _description_. Defaults to None.
        subject (URIRef, optional): _description_. Defaults to None.

    Returns:
        Node (PointCloudNode,MeshNode,GeometryNode,ImageNode)
    """
    #input validation
    if(graphPath and not graph):
            graph = Graph().parse(graphPath)
    if(graph and not subject):
        subject=next(graph.subjects(RDF.type))
    if (subject and graph):    
        nodeType = ut.literal_to_string(graph.value(subject=subject,predicate=RDF.type))
    elif (resource):
        if type(resource) is o3d.geometry.PointCloud:
            nodeType='PointCloudNode'
        elif type(resource) is o3d.geometry.TriangleMesh:
            nodeType='MeshNode'
        elif type(resource) is o3d.geometry:
            nodeType='GeometryNode'
        elif type(resource) is np.ndarray:
            nodeType='ImageNode'        
    else:        
        nodeType = 'Node'

    #node creation
    if 'BIMNode' in nodeType:
        node=BIMNode(graph=graph, graphPath=graphPath, resource=resource,subject=subject, **kwargs)
    elif 'MeshNode' in nodeType:
        node=MeshNode(graph=graph, graphPath=graphPath, resource=resource, subject=subject, **kwargs)
    elif 'GeometryNode' in nodeType:
        node=GeometryNode(graph=graph, graphPath=graphPath, resource=resource, subject=subject, **kwargs)
    elif 'PointCloudNode' in nodeType:
        node=PointCloudNode(graph=graph, graphPath=graphPath, resource=resource, subject=subject, **kwargs)
    elif 'ImageNode' in nodeType:
        node=ImageNode(graph=graph, graphPath=graphPath, resource=resource, subject=subject, **kwargs)
    elif 'SessionNode' in nodeType:
        node=SessionNode(graph=graph, graphPath=graphPath, resource=resource, subject=subject, **kwargs)  
    else:
        node=Node(graph=graph, graphPath=graphPath, resource=resource, subject=subject, **kwargs) 
    return node

def get_linked_nodes(node: Node ,graph:Graph, getResource=False, **kwargs) -> List[Node]:
    """Get related nodes based on linkedNodes variable.\n

    Args:
        0. node (Node): source node to evaluate. \n
        1. graph (Graph): Graph that contains the linkedNodes. \n
        2. getResource (bool, optional): Retrieve the reources. Defaults to False.\n

    Returns:
        List[Node]
    """
    warn("This function is depricated use a SessionNode instead")
    nodelist=[]
    if getattr(node,'linkedNodes',None) is not None:  
        for subject in node.linkedNodes:
            if graph.value(subject=subject,predicate=RDF.type) is not None:
                nodelist.append(create_node(graph=graph,subject=subject, getResource=getResource, **kwargs)) 
    return nodelist